/*
 * hilight.c: The functions in this file handle the syntax highlighting for
 *            MicroEMACS v4.00 (patch level 05 and upper).
 *
 * This file is (c) 1999 Thierry Godefroy.
 *
 *  03/12/99    Improved the ASSEMBLER highlighting (GWass is now supported).
 *              HTML highlighting implemented.
 *  16/09/00    Removed the parenthesis after the return statements that make
 *               gcc to complain.
 *  29/09/00    ASSEMBLER highlighting engine rewrote: it is about 10 times
 *               faster now and as68 is now fully supported as well as label
 *               highlighting, cpp directives highlighting and "intelligent"
 *               highlighting of directives and functions (they are highlighted
 *               only when encountered in the proper field, not in a comment
 *               or a label as they used to be (wrongly) before).
 *              SBASIC engine optimized and line number highlighting imple-
 *               mented.
 *              C, CMD and HTML engine tiny optimisation.
 */

#include    <stdio.h>
#include    <string.h>
#include    <ctype.h>
#include    "estruct.h"
#include    "eproto.h"
#include    "edef.h"
#include    "elang.h"
#include    "hilight.h"

#if CHRATTR

#if PROTO
int PASCAL NEAR highlight(int f, int n)
#else
int PASCAL NEAR highlight(f, n)
int f, n;
#endif
{
    if ((curbp->b_mode & MDHILI) == 0)
        return FALSE;

    hlupdlp = NULL;
    (*curbp->hilight)(TRUE);
    return TRUE;
}

#if PROTO
static VOID PASCAL NEAR c_checkword(LINE *lp, int index, char *word)
#else
static VOID PASCAL NEAR c_checkword(lp, index, word)
LINE *lp;
int index;
char *word;
#endif
{
    register int i, j;

    for (i = 0; i < C_CONTROLS; i++)
        if (!strcmp(word, c_control[i])) {
            for (j = strlen(c_control[i]); j > 0; j--)
                lputa(lp, index-j, hilight1);
            return;
         }
    for (i = 0; i < C_DECLARATIONS; i++)
        if (!strcmp(word, c_declare[i])) {
            for (j = strlen(c_declare[i]); j > 0; j--)
                lputa(lp, index-j, hilight2);
            return;
        }
}

#if PROTO
static VOID PASCAL NEAR c_findcommentstart(LINE *curlp, int index)
#else
static VOID PASCAL NEAR c_findcommentstart(curlp, index)
LINE *curlp;
int index;
char *word;
#endif
{
    register int i, j;
    LINE *lp;

    lp = curlp;

    for (i = index; i >= 0; i--)
        lputa(lp, i, hilight4);

    while ((lp = lback(lp)) != curbp->b_linep) {
        for (i = lp->l_used; i >= 0; i--) {
            lputa(lp, i, hilight4);
            if (i > 0 && lgetc(lp, i) == '*' && lgetc(lp, i-1) == '/') {
                lputa(lp, --i, hilight4);
                return;
            }
        }
    }
}

#if PROTO
static VOID PASCAL NEAR c_engine(LINE *lp)
#else
static VOID PASCAL NEAR c_engine(lp)
LINE * lp;
#endif
{
    char c, cb, ca, word[MAXWORDLENGTH+1];
    register int i, j, n;
    char quote_flag     = FALSE;
    char comment_flag   = FALSE;

    if ((n = lp->l_used) == 0)
        return;

    if (lgetc(lp, 0) == '#') {          /* Preprocessor directive */
        for (i = 0; i < n; i++) {
            c = lgetc(lp, i);
            switch (c) {
                case '"':
                    if (comment_flag)
                        lputa(lp, i, hilight4);
                    else {
                        /* Get the character before and after */
                        cb = ca = 0;
                        if (i > 0)
                            cb = lgetc(lp, i-1);
                        if (i < n-1)
                            ca = lgetc(lp, i+1);
                        /* Now check to see if it is actually a quote */
                        if ((cb != '\'' || ca != '\'') && cb != '\\')
                                quote_flag = !quote_flag;
                        /* Preprocessor directive highlight */
                        lputa(lp, i, hilight3);
                    }
                    break;
                case '/':
                    if (!quote_flag) {
                        if (i < n-1 && lgetc(lp, i+1) == '*') {
                            comment_flag = TRUE;
                            lputa(lp, i, hilight4);
                            break;
                        }
                        if (i > 0 && lgetc(lp, i-1) == '*') {
                            comment_flag = FALSE;
                            lputa(lp, i, hilight4);
                            break;
                        }
                    }
                    if (comment_flag)
                        lputa(lp, i, hilight4);
                    else
                        lputa(lp, i, hilight3);
                    break;
                default:
                    if (comment_flag)
                        lputa(lp, i, hilight4);
                    else
                        lputa(lp, i, hilight3);
            }
        }
    } else {
        j = 0;
        for (i = 0; i < n; i++) {
            c = lgetc(lp, i);
            switch (c) {
                case '"':
                    if (comment_flag)
                        lputa(lp, i, hilight4);
                    else {
                        /* Get the character before and after */
                        cb = ca = 0;
                        if (i > 0)
                            cb = lgetc(lp, i-1);
                        if (i < n-1)
                            ca = lgetc(lp, i+1);
                        /* Now check to see if it is actually a quote */
                        if ((cb != '\'' || ca != '\'') && cb != '\\')
                                quote_flag = !quote_flag;
                        /* no highlight */
                        lputa(lp, i, ATNONE);
                    }
                    j = 0;
                    break;
                case '/':
                    if (!quote_flag) {
                        if (i < n-1 && lgetc(lp, i+1) == '*') {
                            comment_flag = TRUE;
                            lputa(lp, i, hilight4);
                            j = 0;
                            break;
                        }
                        if (i > 0 && lgetc(lp, i-1) == '*') {
                            if (comment_flag) {
                                comment_flag = FALSE;
                                lputa(lp, i, hilight4);
                            } else
                                c_findcommentstart(lp, i);

                            j = 0;
                            break;
                        }
                    }
                    if (comment_flag)
                        lputa(lp, i, hilight4);
                    else
                        lputa(lp, i, ATNONE);
                    j = 0;
                    break;
                default:
                    if (comment_flag) {
                        lputa(lp, i, hilight4);
                        j = 0;
                    } else if (quote_flag) {
                        lputa(lp, i, ATNONE);
                        j = 0;
                    } else {
                        lputa(lp, i, ATNONE);
                        if (isalpha(c) || isdigit(c) || c =='_') {
                            if (j <= MAXWORDLENGTH)
                                word[j++] = c;
                            else
                                j = 0;

                            if (i+1 < n)
                                break;
                            else
                                i++;
                        }
                        word[j] = 0;
                        c_checkword(lp, i, word);
                        j = 0;
                    }
            }
        }
    }
}

#if PROTO
static VOID PASCAL NEAR c_hilite(char flag)
#else
static VOID PASCAL NEAR c_hilite(flag)
char flag;
#endif
{
    register LINE * lp;

    if (!hlupdate && !flag)
        return;

    if (hlupdlp == NULL)
        lp = lforw(curbp->b_linep);
    else
        lp = hlupdlp;

    if (flag == TRUE) {
        mlwrite(TEXT253);
/*              "Highlighting..." */
        while (lp != curbp->b_linep) {
            c_engine(lp);
            lp = lforw(lp);
        }
        mlerase();
    } else
        c_engine(lp);

    hlupdate = FALSE;
    hlupdlp = NULL;
}

#if PROTO
static VOID PASCAL NEAR asm_checkdirective(LINE *lp, int index, char *word)
#else
static VOID PASCAL NEAR asm_checkdirective(lp, index, word)
LINE *lp;
int index;
char *word;
#endif
{
    register int i, j;

    mkupper(word);

    if (word[0] == '.') {
        for (i = 0; i < AS68_DIRECTIVES; i++)
            if (!strcmp(word, asm_directive[i])) {
                for (j = strlen(as68_directive[i]); j > 0; j--)
                    lputa(lp, index-j, hilight1);
                return;
             }
    } else {
        for (i = 0; i < ASM_DIRECTIVES; i++)
            if (!strcmp(word, asm_directive[i])) {
                for (j = strlen(asm_directive[i]); j > 0; j--)
                    lputa(lp, index-j, hilight1);
                return;
             }
    }
}

#if PROTO
static VOID PASCAL NEAR asm_checkfunction(LINE *lp, int index, char *word)
#else
static VOID PASCAL NEAR asm_checkfunction(lp, index, word)
LINE *lp;
int index;
char *word;
#endif
{
    register int i, j;

    mkupper(word);

    for (i = 0; i < ASM_FUNCTIONS; i++)
        if (!strcmp(word, asm_function[i])) {
            for (j = strlen(asm_function[i]); j > 0; j--)
                lputa(lp, index-j, hilight3);
            return;
         }
}

#if PROTO
static VOID PASCAL NEAR asm_engine(LINE *lp)
#else
static VOID PASCAL NEAR asm_engine(lp)
LINE * lp;
#endif
{
    char c, quote, word[MAXWORDLENGTH+1];
    register int i, j, n;
    int k;

    if ((n = lp->l_used) == 0)
        return;
    
    i = 0;
    c = lgetc(lp, 0);

    /* Check for a CPP directive */
    if (c == '#') {
        c_engine(lp);   /* We already know how to highlight this... */
        return;
    }

    /* Check for a comment line */
    if (c == '*' || c == ';' || c == '!') {
asm_comment:
        for (j = i; j < n; j++)
            lputa(lp, j, hilight4);
        return;
    }

    /*
     * The line is not a comment neither a pre-processor directive
     * and we are therefore into a label field.
     */
    
    /* Check for a label, if present highlight it. */
    if (c != ' ' && c != '\t') {
        /* Highlight as a label until we reach a spacing character. */
        do {
            lputa(lp, i++, hilight2);
            if (i == n)
                return;
            c = lgetc(lp, i);
        } while (c != ' ' && c != '\t');
    }
    
    /* Skip spacing characters... */
    while (c == ' ' || c == '\t') {
        lputa(lp, i, ATNONE);
        if (++i == n)
            return;
        c = lgetc(lp, i);
    }

    /*
     * Now we are into the directive/instruction field of the line.
     */

    /* Check again for a comment */
    if (c == '*' || c == ';' || c == '!')
        goto asm_comment;

    j = 0;
    do {
        lputa(lp, i, ATNONE);
        word[j++] = c;
        if (++i == n || j == MAXWORDLENGTH+1)
            break;
        c = lgetc(lp, i);
    } while (c != ' ' && c != '\t');
    word[j] = 0;        
    asm_checkdirective(lp, i, word);  	/* Is this word a directive ? */

    if (i == n)
        return;

    /* If the word was ended because of the MAXWORDLENGTH limit, we
     * must skip the remaining non-spacing characters...
     */
    while (c != ' ' && c != '\t') {
        if (++i == n)
            return;
        c = lgetc(lp, i);
        lputa(lp, i, ATNONE);
    }

    /* Skip the spacing characters up to the next field */
    while (c == ' ' || c == '\t') {
        lputa(lp, i, ATNONE);
        if (++i == n)
            return;
        c = lgetc(lp, i);
    }

    /*
     * At this point we passed the directive/instruction field and
     * and we now enter either the operand field or the comment field.
     */
     
    /* Check again for a comment */
    if (c == ';' || c == '!')
        goto asm_comment;

    /* If this is not a comment, then this is perhaps an operand field
     * and we must therefore search for function names in it until we
     * reach a comment... or the end of the line.
     */
    k = i;
    j = 0;
    for (i = k; i < n; i++) {
        c = lgetc(lp, i);
        switch (c) {
            case '\'':              /* quote: ' */
            case '"':               /* quote: " */
                quote = c;
                do {
                    lputa(lp, i, ATNONE);
                    if (++i == n)
                        return;
                } while ((c = lgetc(lp, i)) != quote);
                lputa(lp, i, ATNONE);
                j = 0;
                break;
            case ';':
            case '!':
                goto asm_comment;
            default:
                lputa(lp, i, ATNONE);
                if (isalpha(c) || isdigit(c)
                    || c =='.' || c == '_' || c == '#') {
                    if (j <= MAXWORDLENGTH)
                        word[j++] = c;
                    else
                        j = 0;
                        
                    if (i+1 != n)
                        break;
                    else
                        i++;
                }
                word[j] = 0;
                asm_checkfunction(lp, i, word);
                j = 0;
        }
    }
}

#if PROTO
static VOID PASCAL NEAR asm_hilite(char flag)
#else
static VOID PASCAL NEAR asm_hilite(flag)
char flag;
#endif
{
    register LINE * lp;

    if (!hlupdate && !flag)
        return;

    if (hlupdlp == NULL)
        lp = lforw(curbp->b_linep);
    else
        lp = hlupdlp;

    if (flag == TRUE) {
        mlwrite(TEXT253);
/*              "Highlighting..." */
        while (lp != curbp->b_linep) {
            asm_engine(lp);
            lp = lforw(lp);
        }
        mlerase();
   } else
        asm_engine(lp);

    hlupdate = FALSE;
    hlupdlp = NULL;
}

#if QDOS

#if PROTO
static VOID PASCAL NEAR bas_checkword(LINE *lp, int index, char *word)
#else
static VOID PASCAL NEAR bas_checkword(lp, index, word)
LINE *lp;
int index;
char *word;
#endif
{
    register int i, j;

    for (i = 0; i < BAS_STRUCTURES; i++)
        if (!strcmp(word, bas_structure[i])) {
            for (j = strlen(bas_structure[i]); j > 0; j--)
                lputa(lp, index-j, hilight1);
            return;
         }
}

#if PROTO
static VOID PASCAL NEAR bas_engine(LINE *lp)
#else
static VOID PASCAL NEAR bas_engine(lp)
LINE * lp;
#endif
{
    char c, ca, quote, word[MAXWORDLENGTH+1];
    register int i, j, n;

    if ((n = lp->l_used) == 0)
        return;

    /* SSB (Structured Super Basic) specific stuff */
    c  = lgetc(lp, 0);
    ca = lgetc(lp, 1);
    /* Check for a "**" or a "##" in first position */
    if ((c == '*' && ca == '*') || (c =='#') && (ca =='#')) {
        for (j = 0; j < n; j++)
            lputa(lp, j, hilight4);
        return;
    /* Check for a pre-processor directive */
    } else if (c == '#') {
        for (j = 0; j < n; j++)
            lputa(lp, j, hilight3);
        return;
    }

    j = 0;
    for (i = 0; i < n; i++) {
        c = lgetc(lp, i);
        /* Check first for a possible line number */
        if (i == 0 && isdigit(c)) {
            do {
                lputa(lp, i, hilight2);
                if (++i == n)
                    return;
                c = lgetc(lp, i);
            } while (isdigit(c));
        }
        switch (c) {
            case '"':               /* quote: " */
            case '\'':              /* quote: ' */
                quote = c;
                do {
                    lputa(lp, i, ATNONE);
                    if (++i == n)
                        return;
                } while ((c = lgetc(lp, i)) != quote);
                lputa(lp, i, ATNONE);
                j = 0;
                break;
            default:
                lputa(lp, i, ATNONE);
                if (isalpha(c) || isdigit(c) || c == '_') {
                    if (j <= MAXWORDLENGTH)
                        word[j++] = c;
                    else
                        j = 0;
                        
                    if (i+1 != n)
                        break;
                    else
                        i++;
                }
                word[j] = 0;
                mkupper(word);
                if (!strcmp(word, "REMARK")) {
                    for (j = -6; i+j < n; j++)
                        lputa(lp, i+j, hilight4);
                    return;
                }
                bas_checkword(lp, i, word);
                j = 0;
        }
    }
}
#if PROTO
static VOID PASCAL NEAR bas_hilite(char flag)
#else
static VOID PASCAL NEAR bas_hilite(flag)
char flag;
#endif
{
    register LINE * lp;

    if (!hlupdate && !flag)
        return;

    if (hlupdlp == NULL)
        lp = lforw(curbp->b_linep);
    else
        lp = hlupdlp;

    if (flag == TRUE) {
        mlwrite(TEXT253);
/*              "Highlighting..." */
        while (lp != curbp->b_linep) {
            bas_engine(lp);
            lp = lforw(lp);
        }
        mlerase();
    } else
        bas_engine(lp);

    hlupdate = FALSE;
    hlupdlp = NULL;
}

#endif /* QDOS */

#if PROTO
static VOID PASCAL NEAR cmd_checkword(LINE *lp, int index, char *word)
#else
static VOID PASCAL NEAR cmd_checkword(lp, index, word)
LINE *lp;
int index;
char *word;
#endif
{
    register int i, j;

    if(strlen(word) <= 1)
        return;

    if (word[0] != '&') {
        for (i = 0; i < CMD_DIRECTIVES; i++)
            if (!strcmp(word, cmd_directives[i])) {
                for (j = strlen(cmd_directives[i]); j > 0; j--)
                    lputa(lp, index - j, hilight1);
                return;
             }
    } else
        for (j = strlen(word); j > 0; j--)
                lputa(lp, index - j, hilight2);
}

#if PROTO
static VOID PASCAL NEAR cmd_engine(LINE *lp)
#else
static VOID PASCAL NEAR cmd_engine(lp)
LINE * lp;
#endif
{
    char c, cb, word[MAXWORDLENGTH+1];
    register int i, j, n;
    char quote_flag     = FALSE;
    char comment_flag   = FALSE;

    if ((n = lp->l_used) == 0)
        return;

    j = 0;
    for (i = 0; i < n; i++) {
        c = lgetc(lp, i);
        /* Check for a label first */
        if (i == 0 && c == '*' ) {
            do {
                lputa(lp, i, hilight3);
                if (++i == n)
                    return;
                c = lgetc(lp, i);
            } while (c != ' ' && c != '\t');
        }
        switch (c) {
            case '"':                           /* Quotes */
                if (comment_flag)
                    lputa(lp, i, hilight4);
                else {
                    /* Get the character before */
                    cb = 0;
                    if (i > 0)
                        cb = lgetc(lp, i-1);
                    if (c != '~')
                        quote_flag = !quote_flag;
                    lputa(lp, i, ATNONE);
                }
                j = 0;
                break;
            case ';':                           /* Comments */
                if (!quote_flag) {
                    comment_flag = TRUE;
                    lputa(lp, i, hilight4);
                    j = 0;
                    break;
                }
                if (comment_flag)
                    lputa(lp, i, hilight4);
                else
                    lputa(lp, i, ATNONE);
                j = 0;
                break;
            default:
                if (comment_flag) {
                    lputa(lp, i, hilight4);
                    j = 0;
                } else if (quote_flag) {
                    lputa(lp, i, ATNONE);
                    j = 0;
                } else {
                    lputa(lp, i, ATNONE);
                    if (isalpha(c) || isdigit(c) || c =='_' || c == '-' ||
                        c == '$' || c == '&' || c == '!' || c == '%') {
                        if (j <= MAXWORDLENGTH)
                            word[j++] = c;
                        else
                            j = 0;

                        if (i+1 != n)
                            break;
                        else
                            i++;
                    }
                    word[j] = 0;
                    cmd_checkword(lp, i, word);
                    j = 0;
                }
        }
    }
}

#if PROTO
static VOID PASCAL NEAR cmd_hilite(char flag)
#else
static VOID PASCAL NEAR cmd_hilite(flag)
char flag;
#endif
{
    register LINE * lp;

    if (!hlupdate && !flag)
        return;

    if (hlupdlp == NULL)
        lp = lforw(curbp->b_linep);
    else
        lp = hlupdlp;

    if (flag == TRUE) {
        mlwrite(TEXT253);
/*              "Highlighting..." */
        while (lp != curbp->b_linep) {
            cmd_engine(lp);
            lp = lforw(lp);
        }
        mlerase();
    } else
        cmd_engine(lp);

    hlupdate = FALSE;
    hlupdlp = NULL;
}

#if PROTO
static VOID PASCAL NEAR html_checkword(LINE *lp, int index, char *word)
#else
static VOID PASCAL NEAR html_checkword(lp, index, word)
LINE *lp;
int index;
char *word;
#endif
{
    register int i;

    if (word[0] == '<') {
        /* The word is a HTML tag */
        for (i = strlen(word); i > 0; i--)
            lputa(lp, index-i, hilight1);
        return;
    }

    if (word[0] == '=')
        /* The word is a parameter value into a HTML tag */
        for (i = strlen(word); i > 0; i--)
            lputa(lp, index-i, hilight2);

    if (index-strlen(word) <= 0 || index == lp->l_used)
        return;
    if (lgetc(lp, index-strlen(word)-1) != '&' || lgetc(lp, index) != ';')
    	return;
    /* The word is a special character &chr_name; */
    for (i = strlen(word)+1; i >= 0; i--)
        lputa(lp, index-i, hilight3);
}

#if PROTO
static VOID PASCAL NEAR html_findcommentstart(LINE *curlp, int index)
#else
static VOID PASCAL NEAR html_findcommentstart(curlp, index)
LINE *curlp;
int index;
char *word;
#endif
{
    register int i, j;
    LINE *lp;

    lp = curlp;

    for (i = index; i >= 0; i--)
        lputa(lp, i, hilight4);

    while ((lp = lback(lp)) != curbp->b_linep) {
        for (i = lp->l_used-1; i >= 0; i--) {
            lputa(lp, i, hilight4);
            if (i < lp->l_used-3 && lgetc(lp, i) == '<' &&
                lgetc(lp, i+1) == '!' && lgetc(lp, i+2) == '-' &&
                lgetc(lp, i+3) == '-') {
                return;
            }
        }
    }
}

#if PROTO
static VOID PASCAL NEAR html_engine(LINE *lp)
#else
static VOID PASCAL NEAR html_engine(lp)
LINE * lp;
#endif
{
    char c, cb, ca, word[MAXWORDLENGTH+1];
    register int i, j, n;
    char quote1_flag    = FALSE;
    char quote2_flag    = FALSE;
    char comment_flag   = FALSE;
    char tag_flag       = FALSE;

    if ((n = lp->l_used) == 0)
        return;

    j = 0;
    for (i = 0; i < n; i++) {
        c = lgetc(lp, i);
        switch (c) {
            case '"':               /* quote: " */
                if (comment_flag)
                    lputa(lp, i, hilight4);
                else if (!quote2_flag && tag_flag) {
                    quote1_flag = !quote1_flag;
                    lputa(lp, i, hilight2);
                } else
                    lputa(lp, i, ATNONE);
                j = 0;
                break;
            case '\'':              /* quote: ' */
                if (comment_flag)
                    lputa(lp, i, hilight4);
                else if (!quote1_flag && tag_flag) {
                    quote2_flag = !quote2_flag;
                    lputa(lp, i, hilight2);
                } else
                    lputa(lp, i, ATNONE);
                j = 0;
                break;
            case '<':
                if (!quote1_flag && !quote2_flag) {
                    if (i < n-3 && lgetc(lp, i+1) == '!' &&
                        lgetc(lp, i+2) == '-' && lgetc(lp, i+3) == '-') {
                        comment_flag = TRUE;
                    }
                }
                if (comment_flag) {
                    lputa(lp, i, hilight4);
                    j = 0;
                } else if (quote1_flag || quote2_flag) {
                    lputa(lp, i, hilight2);
                    j = 0;
                } else {
                    tag_flag = TRUE;
                    lputa(lp, i, hilight1);
                    j = 0;
                    word[j++] = c;
                }
                break;
            case '>':
                if (!quote1_flag && !quote2_flag) {
                    if (i > 1 && lgetc(lp, i-1) == '-' &&
                        lgetc(lp, i-2) == '-') {
                        if (comment_flag) {
                            comment_flag = FALSE;
                            lputa(lp, i, hilight4);
                        } else {
                            html_findcommentstart(lp, i);
                        }
                        j = 0;
                        break;
                    }
                }
                if (comment_flag) {
                    lputa(lp, i, hilight4);
                } else if ((quote1_flag || quote2_flag) && tag_flag) {
                    lputa(lp, i, hilight2);
                } else {
                    tag_flag = quote1_flag = quote2_flag = FALSE;
                    lputa(lp, i, hilight1);
                    word[j] = 0;
                    if (word[0] == '<') {
                        /* The current word is a tag without parameter */
                    	html_checkword(lp, i, word);
                    } else if (j > 0 && i > j) {
                        if (lgetc(lp, i-j-1) == '=') {
                            /* This word is a parameter value in a HTML tag */
                            word[0] = '=';
                            html_checkword(lp, i, word);
                        }
                    }
                }
                j = 0;
                break;
            default:
                if (comment_flag) {
                    lputa(lp, i, hilight4);
                    j = 0;
                } else if (quote1_flag || quote2_flag && tag_flag) {
                    lputa(lp, i, hilight2);
                    j = 0;
                } else {
                    if (tag_flag)
                        lputa(lp, i, hilight3);
                    else
                        lputa(lp, i, ATNONE);
                    if (isalpha(c) || isdigit(c) || c == '/' || c == '!' ||
                        c == '_' || c == '%') {
                        if (j <= MAXWORDLENGTH)
                            word[j++] = c;
                        else
                            j = 0;

                        if (i+1 < n)
                            break;
                        else
                            i++;
                    }
                    word[j] = 0;
                    if (tag_flag && strlen(word) > 0 && i > strlen(word)) {
                        if (lgetc(lp, i-strlen(word)-1) == '=')
                            /* This word is a parameter value in a HTML tag */
                            word[0] = '=';
                    }
                    html_checkword(lp, i, word);
                    j = 0;
                }
        }
    }
}

#if PROTO
static VOID PASCAL NEAR html_hilite(char flag)
#else
static VOID PASCAL NEAR html_hilite(flag)
char flag;
#endif
{
    register LINE * lp;

    if (!hlupdate && !flag)
        return;

    if (hlupdlp == NULL)
        lp = lforw(curbp->b_linep);
    else
        lp = hlupdlp;

    if (flag == TRUE) {
        mlwrite(TEXT253);
/*              "Highlighting..." */
        while (lp != curbp->b_linep) {
            html_engine(lp);
            lp = lforw(lp);
        }
        mlerase();
    } else
        html_engine(lp);

    hlupdate = FALSE;
    hlupdlp = NULL;
}

#if PROTO
VOID PASCAL NEAR custom_hilite(char flag)
#else
VOID PASCAL NEAR custom_hilite(flag)
char flag;
#endif
{
    if (!hlupdate && !flag)
        return;

    execkey(&shlhook, flag, hlupdlp != NULL);

    hlupdate = FALSE;
    hlupdlp = NULL;
}

#if PROTO
char *PASCAL NEAR currentsyntax(VOID)
#else
char *PASCAL NEAR currentsyntax()
#endif
{
    static char result[24];

    if ((curbp->b_mode & MDHILI) == 0) {
        strcpy(result, "DEFAULT");
        return result;
    }

    strcpy(result, "Bug in highlighting !");

    if (curbp->hilight == c_hilite)
        strcpy(result, "C");
    else if (curbp->hilight == asm_hilite)
        strcpy(result, "ASSEMBLER");
#if QDOS
    else if (curbp->hilight == bas_hilite)
        strcpy(result, "SBASIC");
#endif
    else if (curbp->hilight == cmd_hilite)
        strcpy(result, "MACRO");
    else if (curbp->hilight == html_hilite)
        strcpy(result, "HTML");
    else if (curbp->hilight == custom_hilite)
        strcpy(result, "CUSTOM");
    else if (curbp->hilight == nohilite)
        strcpy(result, "DEFAULT");

    return result;
}

#if PROTO
VOID PASCAL NEAR syntaxchoice(char *type)
#else
VOID PASCAL NEAR syntaxchoice(type)
char *type;
#endif
{
    char result[NSTRING+1];

    if ((curbp->b_mode & MDHILI) == 0)
        return;

    strcpy(result, type);
    mkupper(result);

    if (!strcmp(result, "C"))
        curbp->hilight = c_hilite;
    else if (!strcmp(result, "ASSEMBLER"))
        curbp->hilight = asm_hilite;
#if QDOS
    else if (!strcmp(result, "SBASIC"))
        curbp->hilight = bas_hilite;
#endif
    else if (!strcmp(result, "MACRO"))
        curbp->hilight = cmd_hilite;
    else if (!strcmp(result, "HTML"))
        curbp->hilight = html_hilite;
    else if (!strcmp(result, "CUSTOM"))
        curbp->hilight = custom_hilite;
    else if (!strcmp(result, "DEFAULT"))
        curbp->hilight = nohilite;

    hlupdlp = NULL;
    (*curbp->hilight)(TRUE);
}

#if PROTO
VOID PASCAL NEAR nohilite(char flag)
#else
VOID PASCAL NEAR nohilite(flag)
char flag;
#endif
{
    register char   *p, *f;
    char            ext[5] = "";

    if ((curbp->b_mode & MDHILI) == 0)
        return;

    f = curbp->b_fname;

#if QDOS
    if (strlen(f) != 0) {
        p = strrstr(f, "_");
        if ((p != NULL) && (strlen(p) <= 5))
            strcpy(ext, ++p);
        else {
            p = strrstr(f, ".");
            if ((p != NULL) && (strlen(p) <= 5))
                strcpy(ext, ++p);
        }
    }
#else
    if (strlen(f) != 0) {
        p = strrstr(f, ".");
        if ((p != NULL) && (strlen(p) <= 5))
            strcpy(ext, ++p);
    }
#endif

    if (strlen(ext) == 0)
        return;

    mkupper(ext);

    if (!strcmp(ext, "C") || !strcmp(ext, "H"))
        curbp->hilight = c_hilite;
    else if (!strcmp(ext, "ASM") || !strcmp(ext, "S"))
        curbp->hilight = asm_hilite;
#if QDOS
    else if (!strcmp(ext, "BAS") || !strcmp(ext, "SBAS") || !strcmp(ext, "SSB"))
        curbp->hilight = bas_hilite;
#endif
    else if (!strcmp(ext, "CMD") || !strcmp(ext, "RC"))
        curbp->hilight = cmd_hilite;
    else if (!strcmp(ext, "HTM") || !strcmp(ext, "HTML"))
        curbp->hilight = html_hilite;

    if (curbp->hilight != nohilite)
        (*curbp->hilight)(flag);
}

#endif /* CHRATTR */
