/*
 * qdos_c: Operating specific I/O and Spawning functions for MicroEMACS 3.11-4.00
 *      (c) 1991, 92 Richard Kettlewell - (c) 1993-2000 Thierry Godefroy.
 *
 *              16/11/91        Adapted to QDOS from MSDOS.C.
 *                 12/91        Filename completion routines added.
 *                              External prog routines debugged.
 *              18/01/92        Updated for C68 release 2.
 *              20/03/92        Updated for EMACS 3.11.
 *              10/07/92        Updated for C68 release 3.01.
 *                              A bug removed from shellprog().
 *              18/07/92        Minor tidying.
 *              06/09/92        Filename completion routines rewritten.
 *                              Environment support enhanced.
 *
 *      Modifications and additions by Thierry Godefroy:
 *
 *              23/03/93        Updated for C68 release 3.05 (use of system()
 *                               for internal commands execution).
 *              24/03/93        Mode line color configuration added via
 *                               new environment variables EMACS_MDINK and
 *                               EMACS_MDPAPER (see also ql_c & display_c).
 *              03/04/93        - ttopen() & ttclose() moved to ql_c.
 *                              - updates_emacs_vars() moved from ql_c.
 *              08/04/93        Minor debugging of redraws when WINDOW_TEXT=1.
 *              10/04/93        - *getnfile() modified to recognize CST and THOR
 *                                hard disk (sub-)directories.
 *                              - EMACS_MDINK & EMACS_MDPAPER colors are now
 *                                case independant.
 *              28/08/94        Type corrected for qlout (chanid_t).
 *              12/09/98        - Changed default shell name to "sh".
 *                              - Added the "fileinfo" command (see also efunc.h
 *                                ebind.h and eproto.h).
 *              13/09/98        The "fileinfo" command now saves the buffer if
 *                               changed before calling FileInfo II.
 *              22/11/98        Added the "@" prefix to the command line passed
 *                               by shellprog() when "sh" (Adrian Ives' "The
 *                               Shell") is the shell.
 *              03/12/98        Adapted to MicroEMACS v4.00.
 *              10/12/98        Added mlerase() before calling the shell in
 *                               shellprog().
 *              13/12/98        Added the "EMACS_POINTER" environment variable
 *                               (accepts "hand" and "cursor" values).
 *              16/12/98        Useless "exit" command removed in "execprog()".
 *                              Added the scrap stuff.
 *                              fileinfo() now accepts an argument which is used
 *                               as the FileInfo II key.
 *                              Prototyping of all functions.
 *              18/12/98        qexec() function implemented as a work around
 *                               for the buggy fork/exec/qfork functions. The
 *                               workaround is enabled when -DFORKEXECBUG is
 *                               passed to "cc" at compile time.
 *                              "rval" variable is now properly set by all
 *                               spawning functions.
 *              21/12/98        Data directory prompts made conditional to the
 *                               "PROMPT" define.
 *                              Completion routines compilation made conditional
 *                               to the "COMPLETION" define.
 *              22/12/98        The environment variables are now inherited by
 *                               the child jobs which are set up by qexec().
 *              24/12/98        Added the insertscrap() and regionscrap()
 *                               functions.
 *              30/12/98        The "fileinfo" command is no more available when
 *                               in restricted mode.
 *              31/12/98        Corrected a typo into the "dir_transform" struc-
 *                               ture that was the reason for c68 to generate
 *                               warnings.
 *              01/01/99        Added the "EMACS_CTRLROW" environment variable.
 *              07/01/99        Added the "EMACS_LINES" and "EMACS_COLS" envi-
 *                               ronments variables.
 *              16/01/99        Added the spell checking functions.
 *              27/01/99        Added the command line ink and paper.
 *              08/02/99        Added the printbuffer() function.
 *                              Added the EMACS_PRINTER environment variable.
 *              13/02/99        Improved the printbuffer() function.
 *              14/02/99        Improved the spellopen() function (now accepts
 *                               an optional dictionary name) and spellclear()
 *                               function (better error reporting).
 *              19/02/99        Added the IBM PC-8 character translation to
 *                               printbuffer().
 *              20/02/99        Added the printinit() and printcode() functions.
 *              02/03/99        Corrected a bug in split_qdos_flinemane() that
 *                               could crash 68000 based systems (ok for 680x0
 *                               with x>1).
 *              12/03/99        Added the qmenuchar() and qmenuitem() functions.
 *              16/03/99        Added the qmenulist(), qmenuerror(), qmenudir(),
 *                               qmenuext(), qmenufile(), qmenuferror() and
 *                               qmenustring() functions.
 *              21/03/99        Improved the qmenufile() function.
 *                              First attempt for a SUGGEST thing support
 *                               (unsuccessful: bugs in SUGGEST v2.03): new
 *                               spellsuggest() function.
 *              28/03/99        spellsuggest() function fully implemented but
 *                               still useless because of SUGGEST v2.03 bugs.
 *              29/03/99        Some spellsuggest() debugging (but still no
 *                               joy !).
 *              04/04/99        Removed "magic numbers" from qmenu*() functions
 *                               (added QMENU_STR_LENGTH and QMENU_REPLY_STR
 *                               defines; Jochen, if you read this, why are
 *                               QMenu's parameter strings limited to 39
 *                               characters ?).
 *              06/04/99        Added a custom spellsuggest() function when
 *                               compiled without SUGGEST v2 support.
 *              19/04/99        Added a check for the word length in the
 *                               spellsuggest() function (when compiled with
 *                               SUGGEST support): avoiding to call SUGGEST
 *                               v2.03- with word less than three characters
 *                               seems to "cure" (in fact "avoid") the memory
 *                               leaking problem.
 *              13/05/99        Added the CSM server related functions.
 *              26/10/99        Added support for FileInfo II v3.40 additional
 *                               parameter.
 *                              Added the selectfname() function.
 *              31/10/99        Bugs corrected in qmenu-* functions (all C
 *                               strings are now initialized with an empty
 *                               string because mlreply() does not initialize
 *                               them when the parameter is an empty string).
 *              18/11/99        Highlighting support added (insertscrap() calls
 *                               the highlighter for a partial update).
 *              20/11/99        As a result of libqmenu_a debugging, the
 *                               qmenudir() function now asks for a default
 *                               directory name.
 *                              Added the character attributes support for
 *                               printbuffer().
 *              23/11/99        Because of an unknown bug in libqmenu_a that
 *                               crashes the Q40, the qmenu*() functions now
 *                               checks for the EMACS_Q40BUG environment
 *                               variable and abort if it is defined (to any
 *                               value).
 *              03/01/00        The libqmenu_a bug is now corrected in v1.05,
 *                               as a result, the EMACS_Q40BUG check has been
 *                               removed.
 *              16/09/00        Changed a couple of things to remove two gcc
 *                               warnings.
 */

#if QDOS

#include        <stdio.h>
#include        <stdlib.h>
#include        <unistd.h>
#include        <string.h>
#include        <time.h>
#include        <fcntl.h>
#include        <ctype.h>
#include        <qdos.h>
#include        <fi2.h>                 /* For FileInfo II support */
#include        <qptr.h>                /* Needed for menu.h below */
#include        <menu.h>                /* For the QMenu and scrap extensions */
#include        <csrvthg.h>             /* For CSM server support */
#include        "estruct.h"
#include        "eproto.h"
#include        "edef.h"
#include        "elang.h"
#ifdef SUGGEST
#include        "suggest.h"             /* For Suggest v2 */
#endif

#define SEARCH_HIGHLIGHT 10

/* Translate the filenames from UNIX to QDOS conventions on open calls */
int (*_Open)(const char *, int, ...) = qopen;

long _stack             = 24576;        /* Plenty of stack space */
long _memincr           = 10240;        /* Big increments to cut fragmentation */
char _prog_name[]       = PROGNAME;
char _version[]         = VERSION;
char *_endmsg           = NULL;         /* No termination message */
long (*_cmdchannels)()  = NULL;         /* Never inherit channels */
long (*_stackchannels)()= NULL;
void (*_consetup)()     = NULL;         /* Always do version independent setup */
int  (*_cmdwildcard)()  = cmdexpand;    /* Expand filenames on command line */
int  (*_conread)()      = NULL;         /* No input translation */
int  (*_conwrite)()     = NULL;         /* No output translation */

#define NPAUSE          25              /* NPAUSE/50 seconds of pause after a
                                           fence matching in CMODE;
                                           defined in ql_c too. */
/*
 * SHELL definitions.
 * e.g. if we say ^X!fred<NL> i.e. spawn fred in a subjob then
 * the result will be SHELL fred
 * Shell name is taken from environment variable SHELL, if that is
 * defined, otherwise shell is called sh.
 */
#define DEFAULTSHELL    "sh"    /* Shell if no SHELL environment variable */
#define LOCALSHELL      0       /* Convert shellprog() to execprog() when set */
#define SHELLEXEC       " "     /* Shell exec command w/spaces either side */
#define PROMPT          1       /* Define as 0 if you do not want the data
                                   directory to be displayed in the command
                                   prompts */

#define COMPLETION      1       /* Define as 0 if you do not want the file com-
                                   pletion feature */

/* Some defines for QMenu support */

#if MENUS
#define MAX_LIST_ITEMS   32 /* Maximum number of items in the item list menu */
#define QMENU_MAIN_COLOR  1 /* Main (primary) colourway */
#define QMENU_LIST_COLOR  3 /* List (secondary) colourway */
#define QMENU_STR_LENGTH 39 /* Maximum length of QMenu strings */
#define QMENU_REPLY_STR  72 /* Maximum length of the reply string */
#endif

/* Some global variables */

extern chanid_t qlout;                          /* Window channel ID */
#if MENUS
extern int appw_pointer;                        /* Pointer type for menu appw */
extern int ctrl_row;                            /* <CTRL> key row number */
extern int win_lines, win_cols;                 /* EMACS window lines & cols */
#endif
char exredraw;                                  /* Flag */
char server_started = 0;                        /* CSM server flag */
static char server_name[33];                    /* CSM server name */
static chanid_t spell_id = -1;                  /* Spell channel Id */

/*
 * There is a serious bug in the _forkexec() routine of C68 v4.24f (and perhaps
 * previous versions as well) that corrupts memory when calling any of the exec
 * and fork (and qfork) functions. The qexec() function below is much more
 * simple but at least it WORKS !
 *
 * If you want to try fork and exec, then do not define FORKEXECBUG at
 * compile time. If you want the "qexec()" fix, then define it instead.
 */

#ifdef FORKEXECBUG
#define ENV_MAGIC 0x4AFC
extern char *_VENV;
int qexec(char * fname, int * channels, char * cmd, jobid_t owner,
          timeout_t timeout);
#endif

/* Forward function declarations */

int execprog(char *cmd, int infd, int errfd, int outfd, int allowfork);
int qlflush(void);

/*
 * Updates emacs variables according to environment variables.
 */

void update_emacs_vars(void)
{
    char *p;
    int  i;

    if (p = getenv("EMACS_PAUSE")) {
        i = atoi(p);
        if (i <= 0)
            i = NPAUSE;
        term.t_pause = i;
    } else
        term.t_pause = NPAUSE;

    if (p = getenv("EMACS_NOBEEP"))
        beep_flag = FALSE;
    else
        beep_flag = TRUE;

    if (p = getenv("EMACS_PITCH"))
        beep_pitch = atoi(p);
    else
        beep_pitch = 6;

    if (p = getenv("EMACS_DUR"))
        beep_dur = atoi(p);
    else
        beep_dur = 1200;

    if (p = getenv( "EMACS_MDINK"))
        mdline_ink = lookup_color(mkupper(p));

    if (p = getenv("EMACS_MDPAPER"))
        mdline_paper = lookup_color(mkupper(p));

    if (p = getenv( "EMACS_CMDINK"))
        cmdline_ink = lookup_color(mkupper(p));

    if (p = getenv("EMACS_CMDPAPER"))
        cmdline_paper = lookup_color(mkupper(p));

    if (p = getenv("EMACS_PRINTER"))
        strcpy(printer_port, p);
    else
        strcpy(printer_port, "par");

#if MENUS
    if (p = getenv("EMACS_POINTER"))
        appw_pointer = (stricmp(p, "CURSOR") != 0);
    else
        appw_pointer = TRUE;

    if (p = getenv("EMACS_CTRLROW"))
        ctrl_row = atoi(p);
    else
        ctrl_row = 2;

    if (p = getenv("EMACS_LINES"))
        win_lines = atoi(p);
    else
        win_lines = -1;

    if (p = getenv("EMACS_COLS"))
        win_cols = atoi(p);
    else
        win_cols = -1;
#endif

    strcpy(os, "QDOS");     /* Set $os variable to a meaningful value */
}

/*
 * Server related functions.
 */

int startserver(int f, int n)
{
    int ret = 0;

    if (server_started == 0) {
        server_name[0] = '\0';

        if (f == TRUE)
            mlreply(TEXT421, server_name, 32);
/*                  "Server name: " */

        if (strlen(server_name) == 0)
            strcpy(server_name, "emacs");           /* Default server name */

        ret = UseSrvThg(SRVR, server_name);
        if (ret == 0)
            server_started = ((f == TRUE && n == 0) ? -1 : 1);
    }

    return(ret == 0);
}

int stopserver(int f, int n)
{
    if (server_started != 0) {
        FreeSrvThg(SRVR, server_name);
        server_started = 0;
    }

    return(TRUE);
}

int pollserver(int f, int n)
{
    int status, ret = 0;
    short rlen;
    char *ptr;

    if (SrvPend(server_name) == 1) {
        ret = SrvPoll(server_name, 0, (void **) &ptr, &rlen);

        if (ret >= 0) {
            status = docmd(ptr);
            update(TRUE);
            free(ptr);
            SrvRply(ret, status ? "OK" : "KO", -1);
        }
    }

    return (ret >= 0);
}

/*
 * Printing functions.
 */

int sendcodes(chanid_t printer, char *string)
{
    int         i, n, ret;
    char        *p, *q, tmp[256];
    char        *codes[] = {
                    "NUL", "SOH", "STX", "EXT", "EOT", "ENQ", "ACK", "BEL",
                    "BS",  "HT",  "LF",  "VT",  "FF",  "CR",  "SO",  "SI",
                    "DLE", "DC1", "DC2", "DC3", "DC4", "NAK", "SYN", "ETB",
                    "CAN", "EM",  "SUB", "ESC", "FS",  "GS",  "RS",  "US"
                };

    p = string;

    while (*p != '\0') {
        q = strstr(p, "|");
        if (q != NULL) {
            strncpy(tmp, p, (q - p));
            tmp[q-p] = '\0';
            p = q + 1;
        } else {
            strcpy(tmp, p);
            p += strlen(p);
        }
        switch (tmp[0]) {
            case '$':
                if (strlen(tmp) > 1) {
                    n = (int) strtol(&tmp[1], NULL, 16);
                    ret = io_sbyte(printer, -1, n);
                    break;
                }
            case '&':
                if (strlen(tmp) > 1) {
                    n = atoi(&tmp[1]);
                    ret = io_sbyte(printer, -1, n);
                    break;
                }
            default:
                n = -1;
                for (i = 0; i < 32; i++)
                    if (!stricmp(codes[i], tmp)) {
                        n = i;
                        break;
                    }
                if (n == -1)
                    ret = io_sstrg(printer, -1, (void *)tmp,
                                   (short)strlen(tmp));
                else
                    ret = io_sbyte(printer, -1, n);
                break;
        }
        if (ret < 0)
            break;
    }
    if (ret < 0)
        mlwrite(TEXT400, ret);
/*              "QDOS error code %d." */

    return(ret >= 0);
}

int printinit(int f, int n)
{
    chanid_t    printer;
    int         ret;

    printer = io_open(printer_port, 0);
    if (printer < 0) {
        mlwrite(TEXT400, printer);
/*              "QDOS error code %d." */
        return(FALSE);
    }

    ret = sendcodes(printer, prtinit);
    
    io_close(printer);

    return(ret);
}

/*
 * Sends a code to the printer.
 */

int printcode(int f, int n)
{
    chanid_t    printer;
    int         ret;

    if (f != TRUE)
        return(FALSE);

    printer = io_open(printer_port, 0);
    if (printer < 0) {
        mlwrite(TEXT400, printer);
/*              "QDOS error code %d." */
        return(FALSE);
    }

    ret = io_sbyte(printer, -1, n);

    io_close(printer);
    return(ret >= 0);
}

/*
 * Prints the current buffer contents.
 */

int printbuffer(int f, int n0)
{
    LINE            *lp;
    int             nline, npage, tab, ret, i, j, k, n, t, start, end;
    chanid_t        printer;
    char            tmp[32];
#if CHRATTR
    char            *pa;
    char            currattr;
    char            attrflag = FALSE;
#endif
    unsigned char   *p;
    unsigned char   translate[256] = {
                        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
                        0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
                        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
                        0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
                        0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
                        0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
                        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
                        0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
                        0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
                        0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
                        0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
                        0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
                        0x9C, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
                        0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E, 0x6F,
                        0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
                        0x78, 0x79, 0x7A, 0x7B, 0x7C, 0x7D, 0x7E, 0x00,
                        0x84, 0x00, 0x86, 0x82, 0x94, 0x00, 0xED, 0x81,
                        0x87, 0xA4, 0x91, 0x00, 0xA0, 0x85, 0x83, 0x89,
                        0x8A, 0x88, 0x8B, 0xA1, 0x8D, 0x8C, 0xA2, 0x95,
                        0x93, 0xA3, 0x97, 0x96, 0xE1, 0x9B, 0x9D, 0x60,
                        0x8E, 0x00, 0x8F, 0x90, 0x99, 0x00, 0x00, 0x9A,
                        0x80, 0xA5, 0x92, 0x00, 0xE0, 0xEB, 0xE9, 0xE2,
                        0xE6, 0xE3, 0xE8, 0xAD, 0xA8, 0x00, 0x00, 0x6F,
                        0xAE, 0xAF, 0xF8, 0xF6, 0x00, 0x00, 0x00, 0x00,
                        0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7,
                        0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
                        0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7,
                        0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
                        0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7,
                        0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
                        0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7,
                        0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
                    };

    if (!printinit(FALSE, 0))
        return(FALSE);

    printer = io_open(printer_port, 0);
    if (printer < 0) {
        mlwrite(TEXT400, printer);
/*              "QDOS error code %d." */
        return(FALSE);
    }

    mlwrite(TEXT413);
/*          "Printing..." */

    if (tabsize > 0)
        tab = tabsize;
    else
        tab = stabsize;

    if (tab <= 0)
        tab = 1;

#if CHRATTR
    attrflag = ((curbp->b_mode & MDHILI) != 0);
#endif

    lp = lforw(curbp->b_linep);     /* start at the first line.   */
    nline = 0;                      /* track the Number of lines. */
    npage = 1;                      /* track the number of pages. */

    while (lp != curbp->b_linep) {
        start = 1;
        end   = page_size;
#if CHRATTR
        currattr = 0;
        sendcodes(printer, underoff);
        sendcodes(printer, boldoff);
        sendcodes(printer, italoff);
#endif
        if (strlen(title) > 0) {
            ret = io_sstrg(printer, -1, (void *)title, (short)strlen(title));
            for (j = 1; j <= 2; j++) {
                if (ret >= 0) {
                    ret = io_sbyte(printer, -1, '\r');
                    if (ret >= 0 && crlf)
                        ret = io_sbyte(printer, -1, '\n');
                }
            }
            if (ret < 0)
                break;
            start += 2;
        }

        if (page_flag)
            end -= 2;

        for (i = start; i <= end; i++) {
            p = (unsigned char *)lp->l_text;
            n = 0;
            for (j = 1; j <= lused(lp); j++) {
#if CHRATTR
                /* Check if the current character attributes changed */
                pa = (char *)(p + lp->l_size);
                if (attrflag && *pa != currattr) {
                    if ((*pa & ATCOLOR) != (currattr & ATCOLOR) ||
                        (*pa & ATBOLD) != (currattr & ATBOLD))
                        sendcodes(printer, \
                          (*pa & (ATCOLOR | ATBOLD)) != 0 ? boldon : boldoff);
                    if ((*pa & ATUNDER) != (currattr & ATUNDER))
                        sendcodes(printer, \
                                  (*pa & ATUNDER) != 0 ? underon : underoff);
                    if ((*pa & ATITALIC) != (currattr & ATITALIC))
                        sendcodes(printer, \
                                  (*pa & ATITALIC) != 0 ? italon : italoff);
                    currattr = *pa;
                }
#endif                
                if (*p == '\t') {
                    t = tab - n % tab;
                    for (k = 1; k <= t; k++) {
                        ret = io_sbyte(printer, -1, ' ');
                        n++;
                        if (ret < 0)
                            break;
                    }
                } else {
                    if (translate) {
                        tmp[0] = '\0';  /* Default: single char translate */
                        tmp[1] = '\b';  /* Backspace for most multiple trans */
                        switch (*p) {
                            case 0x7f:
                                tmp[0] = 0x4f;
                                tmp[2] = 0x63;
                                tmp[3] = '\0';
                                break;
                            case 0x81:
                                tmp[0] = 0x61;
                                tmp[2] = 0x7e;
                                tmp[3] = '\0';
                                break;
                            case 0x85:
                                tmp[0] = 0x6f;
                                tmp[2] = 0x7e;
                                tmp[3] = '\0';
                                break;
                            case 0x8b:
                                tmp[0] = 0x6f;
                                tmp[1] = 0x65;
                                tmp[2] = '\0';
                                n++;
                                break;
                            case 0xa1:
                                tmp[0] = (char) 0x8e;
                                tmp[2] = 0x7e;
                                tmp[3] = '\0';
                                break;
                            case 0xa5:
                                tmp[0] = (char) 0x99;
                                tmp[2] = 0x7e;
                                tmp[3] = '\0';
                                break;
                            case 0xa6:
                                tmp[0] = 0x4f;
                                tmp[2] = 0x2f;
                                tmp[3] = '\0';
                                break;
                            case 0xab:
                                tmp[0] = 0x4f;
                                tmp[1] = 0x45;
                                tmp[2] = '\0';
                                n++;
                                break;
                            case 0xb5:
                                tmp[0] = 0x3f;
                                tmp[2] = (char) 0xa8;
                                tmp[3] = '\0';
                                break;
                            case 0xb6:
                                tmp[0] = 0x53;
                                tmp[2] = 0x6f;
                                tmp[3] = '\0';
                                break;
                            case 0xbc:
                                tmp[0] = 0x3c;
                                tmp[2] = 0x2d;
                                tmp[3] = '\0';
                                break;
                            case 0xbd:
                                tmp[0] = 0x2d;
                                tmp[2] = 0x3e;
                                tmp[3] = '\0';
                                break;
                            case 0xbe:
                                tmp[0] = (char) 0xb3;
                                tmp[2] = 0x5e;
                                tmp[3] = '\0';
                                break;
                            case 0xbf:
                                tmp[0] = (char) 0xb3;
                                tmp[2] = 0x76;
                                tmp[3] = '\0';
                                break;
                            default:
                                ret = io_sbyte(printer, -1, translate[*p]);
                        }
                        if (strlen(tmp) > 0)
                            ret = io_sstrg(printer, -1, (void *)tmp,
                                           (short)strlen(tmp));
                        n++;
                    } else {
                        ret = io_sbyte(printer, -1, *p);
                        n++;
                    }
                }
                if (ret < 0)
                    break;
                p++;
            }
            if (ret >= 0) {
                ret = io_sbyte(printer, -1, '\r');
                if (ret >= 0 && crlf)
                    ret = io_sbyte(printer, -1, '\n');
            }
            if (ret < 0)
                break;
            ++nline;
            lp = lforw(lp);
            if (lp == curbp->b_linep)
                break;
        }
        if (ret < 0)
            break;

        if (++i < end) {
            for (j = i; j <= end; j++) {
                if (ret >= 0) {
                    ret = io_sbyte(printer, -1, '\r');
                    if (ret >= 0 && crlf)
                        ret = io_sbyte(printer, -1, '\n');
                }
            }
            if (ret < 0)
                break;
        }

        if (page_flag) {
#if CHRATTR
            sendcodes(printer, underoff);
            sendcodes(printer, boldoff);
            sendcodes(printer, italoff);
#endif
            itoa(npage, tmp);
            ret = io_sbyte(printer, -1, '\r');
            if (ret >= 0 && crlf)
                ret = io_sbyte(printer, -1, '\n');
            if (ret >= 0)
                ret = io_sstrg(printer, -1, (void *) "Page ", 5);
            if (ret >= 0)
                ret = io_sstrg(printer, -1, (void *)tmp, strlen(tmp));
            if (ret < 0)
                break;
        }

        if (crlf)
            ret = io_sbyte(printer, -1, '\r');
        if (ret >= 0)
            ret = io_sbyte(printer, -1, '\f');
        if (ret < 0)
            break;
        npage++;
    }

    if (ret < 0)
        mlwrite(TEXT400, ret);
/*              "QDOS error code %d." */
    else
        mlwrite(TEXT414, nline);
/*              "%d lines printed."   */

    io_close(printer);
    return(ret >= 0);
}

/*
 * QTyp spell checker routines
 */

int spellopen(int f, int n)
{
    char    dictionary[NFILEN + 6];
    char    tmp[NFILEN];

    strcpy(dictionary, "spell_");

    if (f == TRUE) {
        if (spellclear(FALSE, 0)) {
            mlreply(TEXT411, tmp, NFILEN);
/*                  "Dictionary file: " */

            strcat(dictionary, tmp);
            strcpy(dictfile, tmp);
            spell_id = io_open(dictionary, 2);
        } else
            return(FALSE);
    } else {
        spellclose(FALSE, 0);
        if (strlen(dictfile) > 0)
            strcat(dictionary, dictfile);
        spell_id = io_open(dictionary, 1);
        if (spell_id == ERR_NI)
            spell_id = io_open(dictionary, 2);
    }
    if (spell_id < 0)
        mlwrite(TEXT410);
/*              "QTyp II is not available." */

    return(spell_id >= 0);
}

int spellclose(int f, int n)
{
    if (spell_id >= 0) {
        io_close(spell_id);
        spell_id = -1;
    }
    return(TRUE);
}

int spellclear(int f, int n)
{
    chanid_t dumb;

    spellclose(FALSE, 0);

    dumb = io_open("spell", 3);
    if (dumb >= 0)
        io_close(dumb);

    if (dumb < 0)
        if (dumb == ERR_IU)
            mlwrite(TEXT412);
/*                  "Error: the current dictionary is in use by another job." */
        else
            mlwrite(TEXT400, dumb);
/*                  "QDOS error code %d." */

    return(dumb >= 0);
}

/* SUGGEST v2 routine (it is not implemented as default because of some serious
 *                     bugs in SUGGEST: memory leaking and no ESC mechanism: use
 *                     at your own risks !).
 */

#ifdef SUGGEST

#define VERSINDX(n, v)  (((char *)(&v))[n])
#define STACKSPC        5000

static  struct {
            short l;
            char  s[12];
        } suggest_title = {11, "Suggestions"};

static GLOBALS_t        *globals;
static SUGGEST_HEAD_t   *sugg;

int spellsuggest(int f, int n)
{
    int             ret = -1;
    LQLSTR_t        *word;
    THING_LINKAGE   *link;
    long            d2 = 0, v;
    char            *a2 = NULL;

    if (spell_id < 0)
        if (!spellopen(FALSE, 0))
            return(FALSE);

    /* Find the "suggest" thing */

    if ((long)(sugg = (SUGGEST_HEAD_t *)sms_uthg("suggest", -1, -1, &d2,
        NULL, &v, &link)) < 0)
        return(FALSE);

    /* If "suggest" version too old, abort */

    if (VERSINDX(0, v) < '2')
        goto suggest_free;

    /* If this is SUGGEST v2.03 or earlier, this is a buggy version and we must
     * check for the length of the word to be checked: if it is less than 3
     * characters long, we must refuse to call SUGGEST to avoid important memory
     * leaking.
     */

    if (VERSINDX(0, v) == '2' && VERSINDX(2, v) == '0' && VERSINDX(3, v) <= '3'
        && strlen((char *)pat) <= 2)
        goto suggest_free;

    /* Allocate suggest global area */

    if (((*sugg->alloc_global_P)(&globals, spell_id, NULL, 0, -1, -1)) < 0)
        goto suggest_free;

    /* Set the stack limit */

    if (((*sugg->stacklimit_P)(globals, (char *)&n - STACKSPC + 12)) < 0)
        goto suggest_end;

    /* Get the misspelled word */

    strcpy(globals->word, (char *)pat);

    /* Check the history (this does not seem to work with SUGGEST v2.03...) */

    if ((globals->alltype = (ALLTYPE)(*sugg->check_all_P)(globals->allrec,
        &(globals->allbuf), globals->word, NULL)) > 0) {

        /* If explicit then use it (replace/ignore all) */
        if (globals->alltype == EXPLICIT) {
            if (globals->allbuf.l != 0) {
                delbword(FALSE, 1);
                linstr(globals->allbuf.s);
            }
            ret = 0;
            goto suggest_end;
        }
        if (globals->alltype < 0)
            return(FALSE);
    }

    /* Find suggestions and returns the user's choice (alas SUGGEST v2.03 does
     * not provide any provision to recognize an "ESC" event: it is therefore
     * not possible to interrupt the spell checking unless we consider the
     * "Ignore word" action as equivalent to "ESC"...).
     */

    word = (*sugg->pull_dn_P)(globals, (LQLSTR_t *)&suggest_title);

    spell_id = globals->spell;
    ret      = (int)word;

    if (ret > 0 && word->l != 0) {
        delbword(FALSE, 1);
        linstr(word->s);
    }

suggest_end:
    (*sugg->freemem_P)(globals);

suggest_free:
    sms_fthg("suggest", -1, &d2, 0, NULL, &a2);
    return(ret >= 0);
}

#else

int spellsuggest(int f, int n)
{
    int     ret;
    char    deflt[QMENU_STR_LENGTH+1], buffer[QMENU_STR_LENGTH+1];

    if (spell_id < 0)
        if (!spellopen(FALSE, 0))
            return(FALSE);

    if (strlen((char *)pat) > QMENU_STR_LENGTH) {
        strncpy(deflt, (char *)pat, QMENU_STR_LENGTH);
        deflt[QMENU_STR_LENGTH] = '\0';
    } else
        strcpy(deflt, (char *)pat);

    ret = ItemSelect("Unknown word !", deflt, "Replace", "Ignore", "Learn",
                     -1, -1, QMENU_MAIN_COLOR);

    switch (ret) {
        case 1:
            ret = ReadString("Replace word", "Please enter the correct word:",
                             deflt, QMENU_STR_LENGTH, buffer, -1, -1,
                             QMENU_MAIN_COLOR);
            if (ret >= 0 && strlen(buffer) != 0) {
                delbword(FALSE, 1);
                linstr(buffer);
            } else
                return(FALSE);
            break;
        case 2:
            break;
        case 3:
            fs_save(spell_id, deflt, strlen(deflt));
            break;
        default:
            return(FALSE);
    }

    return(TRUE);
}

#endif /* SUGGEST */

int spellword(int f, int n)
{
    unsigned char word[NPAT], *p;

    if (spell_id < 0)
        if (!spellopen(FALSE, 0))
            return(FALSE);

    if (f && (n < 0))
        backword(FALSE, 1);

    while (!inword())
        if (forwchar(FALSE, 1) == FALSE)
            return(FALSE);

    curwp->w_markp[SEARCH_HIGHLIGHT] = curwp->w_dotp;
    curwp->w_marko[SEARCH_HIGHLIGHT] = curwp->w_doto;

    p = word;
    while (inword()) {
        *p++ = lgetc(curwp->w_dotp, curwp->w_doto);
        if (forwchar(FALSE, 1) == FALSE || (p - word >= NPAT))
            break;
    }
    *p = '\0';

    curwp->w_markp[SEARCH_HIGHLIGHT+1] = curwp->w_dotp;
    curwp->w_marko[SEARCH_HIGHLIGHT+1] = curwp->w_doto;

    if (io_sstrg(spell_id, 0, (void *)word, strlen((char *)word)) >= 0) {
        curwp->w_markp[SEARCH_HIGHLIGHT] = (LINE *)NULL;
        return(TRUE);
    } else {
        strcpy((char *)pat, (char *)word);
        setjtable();
        io_fstrg(spell_id, -1, (void *)word, NPAT-1);
        if (!f && n == 0) {
            update(TRUE);
            return(spellsuggest(FALSE, 0));
        } else
            return(FALSE);
    }
}

int spellcheck(int f, int n)
{
    if (spell_id < 0)
        if (!spellopen(FALSE, 0))
            return(FALSE);

    while (spellword(FALSE, 0)) ;

    return(TRUE);
}

/*
 * QMenu functions:
 */

#if MENUS

int qmenuchar(int f, int n)
{
    char    buffer[2] = "";
    int     ret;

    if (n == 0)
        return(TRUE);

    if (n < 0)
        return(FALSE);

    ret = CharSel("Please select", 0x3f, buffer, NULL, -1, -1,
                   QMENU_MAIN_COLOR, QMENU_LIST_COLOR);

    if (ret == 0)
        return(linsert(n, buffer[0]));
    else
        return(FALSE);
}

int qmenuitem(int f, int n)
{
    char    menu_title[QMENU_STR_LENGTH+1] = "";
    char    menu_prompt[QMENU_STR_LENGTH+1] = "";
    char    item[3][QMENU_STR_LENGTH+1] = {"", "", ""};
    char    tmp1[20] = "";
    char    tmp2[2] = "";
    int     i, ret;

    mlreply(TEXT415, menu_title, QMENU_STR_LENGTH);
/*          "Menu title: " */

    mlreply(TEXT416, menu_prompt, QMENU_STR_LENGTH);
/*          "Menu prompt: " */

    for (i = 0; i <= 2; i++) {
        strcpy(tmp1, TEXT417);
/*              "Menu item n " */
        itoa(i + 1, tmp2);
        strcat(tmp1, tmp2);
        strcat(tmp1, ": ");
        mlreply(tmp1, item[i], QMENU_STR_LENGTH);
        if (strlen(item[i]) == 0)
            break;
    }

    ret = ItemSelect(menu_title, menu_prompt, item[0], item[1], item[2],
                     -1, -1, QMENU_MAIN_COLOR);

    itoa(ret, rval);
    return(ret >= -1);
}

int qmenulist(int f, int n)
{
    int     i, ret;
    char    menu_title[QMENU_STR_LENGTH+1] = "";
    char    items[MAX_LIST_ITEMS][QMENU_STR_LENGTH+1], *p[MAX_LIST_ITEMS];
    char    tmp1[20] = "";
    char    tmp2[3] = "";

    mlreply(TEXT415, menu_title, QMENU_STR_LENGTH);
/*          "Menu title: " */

    for (i = 0; i < MAX_LIST_ITEMS; i++) {
        p[i] = NULL;
        items[i][0] = '\0';
    }

    for (i = 0; i < MAX_LIST_ITEMS; i++) {
        strcpy(tmp1, TEXT417);
/*              "Menu item n " */
        itoa(i, tmp2);
        strcat(tmp1, tmp2);
        strcat(tmp1, ": ");
        p[i] = &items[i][0];
        mlreply(tmp1, items[i], QMENU_STR_LENGTH);
        if (strlen(items[i]) == 0)
            break;
    }

    ret = ListSelect(menu_title, p, i, 4, 0, 0, -1, -1,
                     QMENU_MAIN_COLOR, QMENU_LIST_COLOR, NULL);

    itoa(ret, rval);
    return(ret >= -1);
}

int qmenudir(int f, int n)
{
    char    menu_title[QMENU_STR_LENGTH+1] = "";
    char    buffer[QMENU_STR_LENGTH+1] = "";
    int     ret;

    mlreply(TEXT415, menu_title, QMENU_STR_LENGTH);
/*          "Menu title: " */

    mlreply(TEXT419, buffer, QMENU_STR_LENGTH);
/*          "Default directory: " */

    ret = DirSelect(menu_title, buffer, 0, -1, -1,
                    QMENU_MAIN_COLOR, QMENU_LIST_COLOR);

    if (ret < 0)
        itoa(ret, rval);
    else
        strcpy(rval, buffer);

    return(ret >= 0);
}

int qmenuext(int f, int n)
{
    char    buffer[7] = "";
    int     ret;

    ret = ExtSelect(buffer, -1, -1, QMENU_MAIN_COLOR);

    if (ret < 0)
        itoa(ret, rval);
    else
        strcpy(rval, buffer);

    return(ret >= 0);
}

int qmenufile(int f, int n)
{
    char    menu_title[QMENU_STR_LENGTH+1] = "";
    char    dir[QMENU_STR_LENGTH+1] = "";
    char    ext[7] = "";
    char    buffer[QMENU_STR_LENGTH+1] = "";
    char    *dirp, *extp;
    int     ret;

    mlreply(TEXT415, menu_title, QMENU_STR_LENGTH);
/*          "Menu title: " */

    mlreply(TEXT419, dir, QMENU_STR_LENGTH);
/*          "Default directory: "*/
    if (strlen(dir) != 0)
        dirp = dir;
    else
        dirp = NULL;

    mlreply(TEXT420, ext, 6);
/*          "Default extension: "*/
    if (strlen(ext) != 0)
        extp = ext;
    else
        extp = NULL;

    mlreply(TEXT420, buffer, QMENU_STR_LENGTH);
/*          "Default filename: "*/

    ret = FileSelect(menu_title, dirp, extp, buffer, 0, -1, -1,
                     QMENU_MAIN_COLOR, QMENU_LIST_COLOR);

    if (ret < 0)
        itoa(ret, rval);
    else
        strcpy(rval, buffer);

    return(ret >= 0);
}

int qmenustring(int f, int n)
{
    char    menu_title[QMENU_STR_LENGTH+1] = "";
    char    menu_prompt[QMENU_STR_LENGTH+1] = "";
    char    deflt[QMENU_STR_LENGTH+1] = "";
    char    buffer[QMENU_REPLY_STR+1] = "";
    char    tmp1[20] = "";
    char    tmp2[2] = "";
    int     i, ret;

    mlreply(TEXT415, menu_title, QMENU_STR_LENGTH);
/*          "Menu title: " */

    mlreply(TEXT416, menu_prompt, QMENU_STR_LENGTH);
/*          "Menu prompt: " */

    mlreply(TEXT418, deflt, QMENU_STR_LENGTH);
/*          "Default value: "*/

    ret = ReadString(menu_title, menu_prompt, deflt, QMENU_REPLY_STR, buffer,
                     -1, -1, QMENU_MAIN_COLOR);

    if (ret < 0)
        itoa(ret, rval);
    else
        strcpy(rval, buffer);

    return(ret >= -1);
}

int qmenuerror(int f, int n)
{
    int ret;

    if (!f)
        return(FALSE);

    ret = ReportError(n, -1, -1, QMENU_MAIN_COLOR);
    itoa(ret, rval);

    return(ret >= 0);
}

int qmenuferror(int f, int n)
{
    int ret;

    if (!f)
        return(FALSE);

    ret = FileError(n, 1, -1, -1, QMENU_MAIN_COLOR);
    itoa(ret, rval);

    return(ret >= 0);
}

#endif

/*
 * Scrap manipulation routines:
 */

/* Copy scrap to kill buffer */

long copy_from_scrap(STHG *scrap, SUDATA *u)
{
    int     i;
    char    c;
    char *  p;

    if(scrap->type == 0)    /* Paranoia: in case data type JUST changed... */
    {
        p = scrap->data + 2;
        for (i = 1; i <= scrap->cursize - 2; i++) {
            c = *p++;
            if (c == '\n')
                c = '\r';
            if (!kinsert(FORWARD, c)) {
                return(ERR_OM);
            }
        }
    } else {
        return(ERR_BP);
    }
    return(0);
}

/*
 * This routine copies the scrap contents (provided the scrap holds text)
 * into the kill buffer. When passed a non-zero parameter, the scrap contents
 * is appended to the existing kill buffer. If there is no parameter or if the
 * parameter is 0, then the old kill buffer is overwritten by the scrap
 * contents.
 */

int scrapkill(int f, int n)
{
    unsigned char   datatype;
    unsigned short  modifycount;
    unsigned long   buflen;
    int             ret;

    ret = GetScrapInfo(&datatype, &modifycount, &buflen);
    if (ret != 0) {
        mlwrite(TEXT406);
/*             "Error: Scrap extension not loaded." */
        goto scrapkillend;
    }

    if (datatype != 0) {
        mlwrite(TEXT405);
/*             "Error: the scrap contents is not of the text type." */
        ret = ERR_BP;
        goto scrapkillend;
    }

    if (!f || (f && n == 0))
        clear_ring(FALSE, 0);

    ret = GetScrap(0, NULL, 0, copy_from_scrap);

    if (ret != 0)
        mlwrite(TEXT400, ret);
/*              "QDOS error code %d." */
    else
        mlwrite(TEXT408);

scrapkillend:
    return(ret == 0);
}

/*
 * This routine copies the kill buffer into the scrap. When passed a non-zero
 * parameter, the kill buffer contents is appended to the existing scrap (if
 * the scrap holds text data). If there is no parameter or if the parameter is
 * 0, then the old scrap contents is overwritten by the new data (whatever data
 * type was in the scrap).
 */

int killscrap(int f, int n)
{
    QD_text_t *         dataptr;
    unsigned char       datatype;
    unsigned short      modifycount;
    unsigned long       buflen;
    int                 size   = 0;
    char *              ptr;
    char *              sp;
    int                 ret, counter;
    KILL *              kptr;               /* pointer into kill buffer */

    /* Make sure there is something into the kill buffer */
    if (kbufh[kill_index] == NULL) {
        ret = 0;                            /* not an error, just nothing */
        goto killscrapend;
    }

    ret = GetScrapInfo(&datatype, &modifycount, &buflen);
    if (ret != 0) {
        mlwrite(TEXT406);
/*             "Error: Scrap extension not loaded." */
        goto killscrapend;
    }

    /*
     * If we want to append to existing scrap contents, make sure that the
     * existing data type is correct and that there is enough room...
     */
    if (f && n != 0) {
        if (datatype != 0) {
            mlwrite(TEXT405);
/*                 "Error: the scrap contents is not of the text type." */
            ret = ERR_BP;
            goto killscrapend;
        }

        buflen = 32767 - buflen;            /* Maximum string length to add */
        if (buflen <= 0) {
            mlwrite(TEXT404);
/*                 "Error: scrap buffer full." */
            ret = ERR_BO;
            goto killscrapend;
        }
    } else
        buflen = 32767;                     /* Maximum string length */

    dataptr = (QD_text_t *) malloc(buflen + 2);
    if (dataptr) {
        ptr = (char *) dataptr->chrs;
        size = 0;

        /* For each time.... */
        while (n--) {
            if (kskip[kill_index] > 0) {
                kptr = kbufh[kill_index];
                sp = &(kptr->d_chunk[kskip[kill_index]]);
                counter = kskip[kill_index];
                while (counter++ < KBLOCK) {
                    if (size >= buflen) {
                        mlwrite(TEXT404);
/*                             "Error: scrap buffer full." */
                        ret = ERR_BO;
                        break;
                    }
                    if (*sp == '\r') {
                        *ptr++ = '\n';
                        sp++;
                    } else
                        *ptr++ = *sp++;
                    size++;
                }
                kptr = kptr->d_next;
            } else {
                kptr = kbufh[kill_index];
            }
        
            if (kptr != (KILL *)NULL) {
                while (kptr != kbufp[kill_index]) {
                    sp = kptr->d_chunk;
                    for(counter = 0; counter < KBLOCK; counter++) {
                        if (size >= buflen) {
                            mlwrite(TEXT404);
/*                                 "Error: scrap buffer full." */
                            ret = ERR_BO;
                            break;
                        }
                        if (*sp == '\r') {
                            *ptr++ = '\n';
                            sp++;
                        } else
                            *ptr++ = *sp++;
                        size++;
                    }
                    kptr = kptr->d_next;
                }
                counter = kused[kill_index];
                sp = kptr->d_chunk;
                while (counter--) {
                    if (size >= buflen) {
                        mlwrite(TEXT404);
/*                             "Error: scrap buffer full." */
                        ret = ERR_BO;
                        break;
                    }
                    if (*sp == '\r') {
                        *ptr++ = '\n';
                        sp++;
                    } else
                        *ptr++ = *sp++;
                    size++;
                }
            }
        }
        if (ret == 0) {
            dataptr->len = size;
            ret = PutScrap(0, (void *) dataptr, (f && n != 0), NULL);
            if (ret != 0)
                mlwrite(TEXT400, ret);
/*                      "QDOS error code %d." */
            else
                mlwrite(TEXT407);
        }
        free(dataptr);
    } else {
        mlwrite(TEXT94);
/*             "%%Out of memory" */
        ret = ERR_OM;
    }

killscrapend:
    return(ret == 0);
}

/*
 * This routine clears the scrap. The parameters and any error are ignored.
 */

int erasescrap(int f, int n)
{
    ClearScrap();
    return(TRUE);
}

/* Insert scrap into the current buffer */

#define Char_insert(a)  (a == '\r' ? lnewline() : linsert(1, a))

long insert_from_scrap(STHG *scrap, SUDATA *u)
{
    int     i;
    char    c;
    char *  p;

    if(scrap->type == 0)    /* Paranoia: in case data type JUST changed... */
    {
        p = scrap->data + 2;
        for (i = 1; i <= scrap->cursize - 2; i++) {
            c = *p++;
            if (c == '\n')
                c = '\r';
            Char_insert(c);
        }
    } else {
        return(ERR_BP);
    }
    return(0);
}

/*
 * This routine inserts the scrap contents (provided the scrap holds text)
 * into the current buffer.
 */

int insertscrap(int f, int n)
{
    unsigned char   datatype;
    unsigned short  modifycount;
    unsigned long   buflen;
    int             ret;

    ret = GetScrapInfo(&datatype, &modifycount, &buflen);
    if (ret != 0) {
        mlwrite(TEXT406);
/*             "Error: Scrap extension not loaded." */
        goto insertscrapend;
    }

    if (datatype != 0) {
        mlwrite(TEXT405);
/*             "Error: the scrap contents is not of the text type." */
        ret = ERR_BP;
        goto insertscrapend;
    }

    ret = GetScrap(0, NULL, 0, insert_from_scrap);

    if (ret != 0)
        mlwrite(TEXT400, ret);
/*              "QDOS error code %d." */
    else {
        mlwrite(TEXT408);
#if CHRATTR
        hlupdlp = NULL;
        (*curbp->hilight)(TRUE);
#endif
    }

insertscrapend:
    return(ret == 0);
}

/*
 * This routine copies the current region to the scrap
 */

int regionscrap(int f, int n)
{
    QD_text_t *         dataptr;
    unsigned char       datatype;
    unsigned short      modifycount;
    unsigned long       buflen;
    int                 size = 0;
    int                 ret;
    char *              ptr;
    register LINE *     linep;
    register int        loffs;
    REGION              region;

    /* Is there any region defined ? */

    if ((getregion(&region)) != TRUE)
        return(FALSE);

    ret = GetScrapInfo(&datatype, &modifycount, &buflen);
    if (ret != 0) {
        mlwrite(TEXT406);
/*             "Error: Scrap extension not loaded." */
        goto regionscrapend;
    }

    /*
     * If we want to append to existing scrap contents, make sure that the
     * existing data type is correct and that there is enough room...
     */
    if (f && n != 0) {
        if (datatype != 0) {
            mlwrite(TEXT405);
/*                 "Error: the scrap contents is not of the text type." */
            ret = ERR_BP;
            goto regionscrapend;
        }

        buflen = 32767 - buflen;            /* Maximum string length to add */
        if (buflen <= 0) {
            mlwrite(TEXT404);
/*                 "Error: scrap buffer full." */
            ret = ERR_BO;
            goto regionscrapend;
        }
    } else
        buflen = 32767;                     /* Maximum string length */

    dataptr = (QD_text_t *) malloc(buflen + 2);
    if (dataptr) {
        ptr = (char *) dataptr->chrs;
        size = 0;
        linep = region.r_linep;             /* Current line.        */
        loffs = region.r_offset;            /* Current offset.      */
        while (region.r_size--) {
            if (loffs == lused(linep)) {    /* End of line.         */
                if (size >= buflen) {
                    mlwrite(TEXT404);
/*                          "Error: scrap buffer full." */
                    ret = ERR_BO;
                    break;
                }
                *ptr++ = '\n';
                size++;
                linep = lforw(linep);
                loffs = 0;
            } else {                        /* Middle of line.      */
                if (size >= buflen) {
                    mlwrite(TEXT404);
/*                          "Error: scrap buffer full." */
                    ret = ERR_BO;
                    break;
                }
                *ptr++ = lgetc(linep, loffs);
                size++;
                ++loffs;
            }
        }
        if (ret == 0) {
            dataptr->len = size;
            ret = PutScrap(0, (void *) dataptr, (f && n != 0), NULL);
            if (ret != 0)
                mlwrite(TEXT400, ret);
/*                      "QDOS error code %d." */
            else
                mlwrite(TEXT409);
/*                      "Region copied to scrap." */
        }
        free(dataptr);
    } else {
        mlwrite(TEXT94);
/*             "%%Out of memory" */
        ret = ERR_OM;
    }

regionscrapend:
    return(ret == 0);
}

/*
 * FileInfo II call. This routine first checks if the current buffer was
 * changed; if so then the buffer is saved. Then it calls FileInfo II, passing
 * it the filename corresponding to the current buffer. When a parameter is
 * passed to this function, it is used as a key for FileInfo II features (read
 * FileInfo II manual for possible key values).
 *
 * Bound to ^X I
 */

int fileinfo(int f, int n)
{
    FI2INFO_t   fi2;
    int         key = -1;
    int         ret;
    char        argument[NFILEN+256];
    char        tmp[256];

    /* Don't allow this command if restricted */
    if (restflag)
        return(resterr());

    if ((ret = fi2_info(&fi2)) != 0) {
        mlwrite(TEXT402);
/*              "Error: FileInfo II is not loaded." */
        goto fileinfoend;
    }

    if (f) {
        /* Parameters were passed... */

        /* Get the key. */
        if (n > fi2.max_key || n < -2) {
            mlwrite(TEXT403);
/*                 "Error: Not a valid FileInfo II key." */
            ret = ERR_BP;
            goto fileinfoend;
        }
        key = n;

        /* Get the name of the file to process. If an empty name is passed
           then use the current buffer filename. */

        mlreply(TEXT422, tmp, NFILEN);
/*              "File to process name (nothing for current file): " */
        if (strlen(tmp) > 0)
            strcpy(argument, tmp);
        else {
            /* There is no filename, so ensure the buffer is saved... */
            if ((curbp->b_flag & BFCHG) != 0) { /* If buffer changed */
                mlwrite(TEXT103, curbp->b_fname);
/*                      "[Saving %s]" */
                if (!filesave(FALSE, 0))
                    return(FALSE);
            }
            /* ... and take the buffer filename as the file to process. */
            strcpy(argument, curbp->b_fname);
        }

        mlreply(TEXT423, tmp, 255);
/*              "Additional parameter: " */
        /* If there is an additionnal parameter, then use it... */
        if (strlen(tmp) > 0) {
            if (strlen(argument) & 1)
                strcat(argument, " ");
            strcat(argument, FI2_PAR_MARKER);
            strcat(argument, tmp);
        }

    } else {
        /* No parameter was passed, use default filename and key */
        if ((curbp->b_flag & BFCHG) != 0) {     /* If buffer changed */
            mlwrite(TEXT103, curbp->b_fname);
/*                  "[Saving %s]" */
            if (!filesave(FALSE, 0))
                return(FALSE);
        }
        strcpy(argument, curbp->b_fname);
    }

    ret = fi2_call(-1, key, argument);
    if (ret != 0)
        mlwrite(TEXT400, ret);
/*              "QDOS error code %d." */

fileinfoend:
    itoa(ret, rval);
    return(ret == 0);
}

int infname(void)
{
    char c;

    /* the end of a line is never in a filename */
    if (curwp->w_doto == lused(curwp->w_dotp))
        return(FALSE);

    /* grab the character to check */
    c = lgetc(curwp->w_dotp, curwp->w_doto);

    if (is_letter(c))
        return(TRUE);
    if (c>='0' && c<='9')
        return(TRUE);
    if (c == '_' || c == '/' || c == '.' || c == '~' || c == '+' || c == '-')
        return(TRUE);

    return(FALSE);
}

/*
 * This function sets mark 0 to the start of the current filename
 * and then go to the end of the filename. The filename is hilighted.
 * If passed an argument > 0, the search pattern is set to this word,
 * if passed an argument < 0, the word is copied into the kill buffer.
 */
int selectfname(int f, int n)
{
    char            flag = TRUE;
    unsigned char * p;

    if (f && (n > 0))
        p = pat;

    while (!infname())
        if (backchar(FALSE, 1) == FALSE)
            return(FALSE);

    while (infname())
        if (backchar(FALSE, 1) == FALSE) {
            flag = FALSE;
            break;
        }

    if (flag)
        forwchar(FALSE, 1);

    curwp->w_markp[0] = curwp->w_dotp;
    curwp->w_marko[0] = curwp->w_doto;
    curwp->w_markp[SEARCH_HIGHLIGHT] = curwp->w_dotp;
    curwp->w_marko[SEARCH_HIGHLIGHT] = curwp->w_doto;

    while (infname()) {
        if (f && (n > 0))
            *p++ = lgetc(curwp->w_dotp, curwp->w_doto);
        if (forwchar(FALSE, 1) == FALSE)
            break;
    }

    curwp->w_markp[SEARCH_HIGHLIGHT+1] = curwp->w_dotp;
    curwp->w_marko[SEARCH_HIGHLIGHT+1] = curwp->w_doto;

    if (f && (n > 0)) {
        *p = '\0';
        setjtable();
        mlwrite(TEXT251, pat);
/*              "Search pattern set to '%s'" */
    }

    if (f && (n < 0))
        copyregion(FALSE, 0);

    return(TRUE);
}

/*
 * Create a subjob with a copy of the command interpreter in it. When the
 * command interpreter exits, mark the screen as garbage so that you do a full
 * repaint. Bound to "^X C".
 *
 * If you say ^X$set SHELL=multib_exe<NL> (note capitals) and the Minerva
 * routine multib_exe is on the PATH or in the program directory then
 * you can have SuperBASIC as a shell from MicroEMACS.
 *
 * A lot of this code was borrowed pretty directly from the MSDOS version.
 */

int spawncli(int f, int n)
{
    int s;

    /* don't allow this command if restricted */
    if (restflag)
        return(resterr());

    movecursor(term.t_nrow, 0);             /* Seek to last line. */
    qlflush();
    TTkclose();
    s = shellprog("");
    TTkopen();
    if (exredraw)
#if WINDOW_TEXT
        refresh_screen(first_screen);
#else
        sgarbf = TRUE;
#endif
    return(s);
}

/*
 * Run a one-liner in a subjob. When the command returns, wait for a single
 * character to be typed, then mark the screen as garbage so a full repaint
 * is done. Bound to "^X !".
 */

int spawn(int f, int n)
{
    int s;
    char line[NLINE];
#if PROMPT
    char dnam[50];
#endif

    /* don't allow this command if restricted */
    if (restflag)
        return(resterr());
#if PROMPT
    getcwd(dnam, sizeof dnam);
    strcat(dnam, "> ");
    if ((s=mlreply(dnam, line, NLINE)) != TRUE)
        return(s);
#else
    if ((s=mlreply("> ", line, NLINE)) != TRUE)
        return(s);
#endif
    movecursor(term.t_nrow, 0);
    TTkclose();
    s = shellprog(line);
    TTkopen();
    /* if we are interactive, pause here */
    if (clexec == FALSE && exredraw) {
        mlputs(TEXT188);    /* "[End]" */
        TTflush();
        tgetc();
    }
    if (exredraw)
#if WINDOW_TEXT
        refresh_screen(first_screen);
#else
        sgarbf = TRUE;
#endif
    return(s);
}

/*
 * Run an external program with arguments. When it returns, wait for a single
 * character to be typed, then mark the screen as garbage so a full repaint is
 * done. Bound to "^X $".
 */

int execprg(int f, int n)
{
    int  s;
    char line[NLINE];
#if PROMPT
    char dnam[50];
#endif
    /* don't allow this command if restricted */
    if (restflag)
        return(resterr());

    /* Get command */
#if PROMPT
    getcwd(dnam, sizeof dnam);
    strcat(dnam, "> ");
    if ((s = mlreply(dnam, line, NLINE)) != TRUE)
        return(s);
#else
    if ((s = mlreply("> ", line, NLINE)) != TRUE)
        return(s);
#endif

    /* Execute command */
    movecursor(term.t_nrow, 0);
    TTkclose();
    s = execprog(line, fileno(stdin), fileno(stderr), fileno(stdout), 1);
    TTkopen();
    /* if we are interactive, pause here */
    if (clexec == FALSE && exredraw) {
        mlputs(TEXT188);    /* "[End]" */
        TTflush();
        tgetc();
    }
    if (exredraw)
#if WINDOW_TEXT
        refresh_screen(first_screen);
#else
        sgarbf = TRUE;
#endif
    return(s);
}

/*
 * Pipe a one line command into a window
 * Bound to ^X @
 */

int pipecmd(int f, int n)
{
    int     s;
    int     outfd;
    EWINDOW *wp;                    /* pointer to new window */
    BUFFER  *bp;                    /* pointer to buffer to zot */
    char    *tmp;                   /* ptr to TMP DOS environment variable */
    char    line[NLINE];            /* command line send to shell */
    char    filnam[NSTRING];        /* name of temporary file */
#if PROMPT
    char    dnam[50];
#endif
    static char bname[] = "filter_out"; /* filename - MUST have an extension */

    /* don't allow this command if restricted */
    if (restflag)
        return(resterr());

    if (tmp = getenv("TMP")) {
        strcpy(filnam, tmp);
        if (filnam[strlen(filnam) - 1] != DIRSEPCHAR)
            strcat(filnam, DIRSEPSTR);
        strcat(filnam, bname);
    } else
        strcpy(filnam, bname);
    mklower(filnam);

    /* Get the command to pipe in */
#if PROMPT
    getcwd(dnam, sizeof dnam);
    strcat(dnam, "> ");
    if ((s = mlreply(dnam, line, NLINE)) != TRUE)
        return(s);
#else
    if ((s = mlreply("> ", line, NLINE)) != TRUE)
        return(s);
#endif

    /* Open the file */
    if((outfd = open(filnam, O_WRONLY | O_CREAT | O_TRUNC)) < 0)
        return(FALSE);

    /* Execute the command */
    movecursor(term.t_nrow, 0);
    TTkclose();
    s = execprog(line, fileno(stdin), fileno(stderr), outfd, 0);
    close(outfd);
    TTkopen();
    if (exredraw)
#if WINDOW_TEXT
        refresh_screen(first_screen);
#else
        sgarbf = TRUE;
#endif

    /* If command failed, return now */
    if (s != TRUE)
        goto tidyexit;

    /* Find the command buffer; it is doesn't exist, create it */
    if (bp = bfind(bname, TRUE, 0)) {
        if (curbp != bp) {              /* Don't mess around if already ours */
            onlywind(FALSE, 1);         /* Make current window full screen */
            if (!(s = splitwind(FALSE, 1)))
                goto tidyexit;
            swbuffer(bp);               /* Switch to our buffer */
        }
        s = readin(filnam, FALSE);      /* Load the text */
        if (!s)
            goto tidyexit;
    } else {                            /* Couldn't find buffer */
        s = FALSE;
        goto tidyexit;
    }

    /* make this window in VIEW mode, update all mode lines */
    curwp->w_bufp->b_mode |= MDVIEW;
    wp = wheadp;
    while (wp != NULL) {
        wp->w_flag |= WFMODE;
        wp = wp->w_wndp;
    }
    s = TRUE;

tidyexit:
    unlink(filnam);
    return(s);
}

/*
 * Filter a buffer through an external program.
 * Bound to ^X #
 */

int filter(int f, int n)
{
    int     s;                      /* return status from CLI */
    BUFFER  *bp;                    /* pointer to buffer to zot */
    char    line[NLINE];            /* command line send to shell */
    char    tmpname[NFILEN];        /* place to store real file name */
    int     infd, outfd;
#if PROMPT
    char    dnam[50];
#endif
    char    *sp;
    char    filnam1[NFILEN], filnam2[NFILEN];
    static char bname1[] = "fltinp";

    /* don't allow this command if restricted */
    if (restflag)
        return(resterr());

    if (curbp->b_mode&MDVIEW)       /* don't allow this command if  */
        return(rdonly());           /*  we are in read only mode.   */

#if PROMPT
    getcwd(dnam, sizeof dnam);
    strcat(dnam, "> ");
    if ((s=mlreply(dnam, line, NLINE)) != TRUE)
        return(s);
#else
    /* get the filter name and its args */
    if ((s=mlreply("> ", line, NLINE)) != TRUE)
        return(s);
#endif

    /* setup the proper file names */
    bp = curbp;
    strcpy(tmpname, bp->b_fname);    /* save the original name */
    strcpy(bp->b_fname, bname1);    /* set it to our new one */

    if (sp = getenv("TMP")) {
        if (sp[strlen(sp) - 1] != DIRSEPCHAR)
            strcat(sp, DIRSEPSTR);
        strcpy(filnam1, sp);
        strcpy(filnam2, sp);
    } else {
        *filnam1 = 0;
        *filnam2 = 0;
    }
    strcat(filnam1, "fltinp_tmp");
    strcat(filnam2, "fltout_tmp");

    /* write it out, checking for errors */
    if (writeout(filnam1, "w") != TRUE) {
        mlwrite(TEXT2);              /* "[Cannot write filter file]" */
        strcpy(bp->b_fname, tmpname);
        return(FALSE);
    }

    /* Open the files */
    if ((infd = open(filnam1, O_RDONLY)) < 0) {
        unlink(filnam1);
        strcpy(bp->b_fname, tmpname);
        return(FALSE);
    }
    if ((outfd = open(filnam2, O_WRONLY | O_CREAT | O_TRUNC)) < 0) {
        close(infd);
        unlink(filnam1);
        strcpy(bp->b_fname, tmpname);
        return(FALSE);
    }

    /* Execute the program */
    movecursor(term.t_nrow, 0);
    TTkclose();
    s = execprog(line, infd, fileno(stderr), outfd, 0);
    close(infd);
    close(outfd);
    TTkopen();
    if (exredraw)
#if WINDOW_TEXT
        refresh_screen(first_screen);
#else
        sgarbf = TRUE;
#endif

    /* on failure, escape gracefully */
    if (s != TRUE || !(s = readin(filnam2, FALSE))) {
        mlwrite(TEXT3);              /* "[Execution failed]" */
        strcpy(bp->b_fname, tmpname);
        unlink(filnam1);
        unlink(filnam2);
        return(s);
    }

    /* reset file name */
    strcpy(bp->b_fname, tmpname);   /* restore name */
    bp->b_flag |= BFCHG;            /* flag it as changed */

    /* and get rid of the temporary file */
    unlink(filnam1);
    unlink(filnam2);
    return(TRUE);
}

/* SHELLPROG: Execute a command in a subshell */

int shellprog(char *cmd)
{
#if LOCALSHELL
    return(*cmd ? execprog(cmd, fileno(stdin), fileno(stderr),
                            fileno(stdout), 1) : FALSE);
#else
    char *shell;                        /* Name of system command processor */
    char comline[NSTRING];              /* constructed command line */

    /* get name of system shell */
    if ((shell = getenv("SHELL")) == NULL)
        shell = DEFAULTSHELL;           /* If no shell variable, use default */

    mkupper(shell);                     /* Make it case independant */

    /* trim leading whitespace off the command */
    while (*cmd == ' ' || *cmd == '\t') /* find out if null command */
        cmd++;

    /* If the command line is not empty, bring up the shell
     * and execute the command.  Otherwise, bring up the
     * shell in interactive mode.
     */

    if (*cmd) {
        strcpy(comline, shell);
        strcat(comline, SHELLEXEC);

        /* If we are to use Adrian Ives' "The Shell" (which filename is "sh"),
         * then we need to prefix the command line with "@" so to force "sh"
         * to execute the command and return immediately.
         */

        if (strrstr(shell, "_SH") || !strcmp(shell, "SH"))
            strcat(comline, "@");

        strcat(comline, cmd);
        mlerase();
        return (execprog(comline, fileno(stdin), fileno(stderr),
                fileno(stdout), 0));
    } else
        mlerase();
        return(execprog(shell, fileno(stdin), fileno(stderr),
               fileno(stdout), 0));
#endif
}

/*
 * EXECPROG: A function to execute a named program with arguments and streams.
 *
 * Given program name FRED, will look for FRED in program directory, data
 * directory, and on PATH environment variable.
 * e.g. if PATH is "ram1_;flp1_system_", prog dir is "flp2_emacs_" and data dir
 * is "flp1_progs_" then the following is the search order:-
 *
 *      FRED
 *      flp2_emacs_FRED
 *      flp1_progs_FRED
 *      ram1_FRED
 *      flp1_system_FRED
 *
 * An & character right at the end will cause a fork instead of an exec.
 *
 * We don't check here for > >> etc; that is left to the called program,
 * although all C68 programs (and progs from many other Ql compilers) support
 * these operators. It's not so difficult for assembly progs to support them
 * either.
 *
 * This is all new.
 */

int execprog(char *cmd, int infd, int errfd, int outfd, int allowfork)
{
    char *sp;                   /* temporary string pointer */
    char prog[NSTRING];         /* program filespec */
    char tail[NSTRING];         /* command tail with length byte */
    int  channels[4];           /* Buffer for channels */
    char forkflag = 0;          /* fork or exec ? */
    char *cmd2 = cmd;           /* saves cmd */
    int  rv;                    /* Numeric return value from subprocess */

    exredraw = 1;
    /* Skip initial spaces */
    while (*cmd == ' ' || *cmd == '\t')
        cmd++;
    /* parse the command name from the command line */
    sp = prog;
    while (*cmd && *cmd != ' ' && *cmd != '\t')
        *sp++ = *cmd++;
    *sp = 0;

    /* and parse out the command tail */
    while (*cmd == ' ' || *cmd == '\t')
        ++cmd;
    strcpy(tail, cmd);

    /*
     * Check for internal commands now - before doing flook(). If CD or CPD are
     * called without args, then the appropriate directory is sent to the
     * output. Of course, this is more use when you are filtering stuff than
     * at other times.
     *
     * SET is new at 10/7/92; if the parameter is a definition of an environment
     * variable (e.g. SET PATH=flp1_;flp2_) then it is passed to putenv();
     * otherwise to getenv(), so that the environment name may be displayed. If
     * there is no parameter then no action it taken.
     *
     * 6/9/1992 SET checks for environment names starting EMACS_ and calls a
     * routine to reset the variables which they map to.
     *
     * EXIT is new at 18/7/92, and only expected to do anything useful in the
     * future.
     *
     * 23/03/93 internal commands now executed via the "system()" call.
     *
     * 16/12/98 useless EXIT command removed.
     */
    if (!stricmp(prog, "CD") || !stricmp(prog, "CPD")) {
        if (*tail)
            exredraw = 0;
        else
            exredraw = (outfd == fileno(stdout));
        rv = system(cmd2);
        return(rv == 0 ? TRUE : FALSE);

    } else if (!stricmp(prog, "SET")) {
        exredraw = (outfd == fileno(stdout));
        system(cmd2);
        rv = 0;
        if (*tail) {
            if (strchr(tail, '=') && !strncmp(tail, "EMACS_", 6)) {
                update_emacs_vars();
                exredraw = 0;
                if (!strncmp(tail, "EMACS_MD", 8)) {
                    upmode();
#if WINDOW_TEXT
                    refresh_screen(first_screen);
#endif
                }
            }
        }
        return(rv == 0 ? TRUE : FALSE);
    }

    /* Look up the program on the path */
    if (!(sp = flook(prog, TRUE)))
        return(FALSE);
    if (sp != prog)
        strcpy(prog, sp);
    if(allowfork) {
        /* Look for a & character, which implies fork instead of exec. */
        sp = tail+strlen(tail);
        if (sp > tail && sp[-1] == '&') {
            exredraw = 0;
            *--sp = 0;
            forkflag = 1;
        }
    }

    sd_setpa(qlout, -1, 0);                 /* Reset colours */
    sd_setst(qlout, -1, 0);
    sd_setin(qlout, -1, 7);

    if (forkflag) {
        /* Forked programs don't get any channels given to them
         * because we would never know when to close them.
         */
#ifdef FORKEXECBUG
        rv = qexec(prog, NULL, tail, -1, 0);
#else
        rv = forklp(prog, NULL, prog, tail, NULL);
        if (rv > 65535)                     /* Don't return job IDs! */
            rv = 0;
#endif
    } else {
        channels[0] = 3;                    /* 3 channels */
        channels[1] = infd;                 /* Input file */
        channels[2] = errfd;                /* Error file (usually screen) */
        channels[3] = outfd;                /* Output file */

#ifdef FORKEXECBUG
        rv = qexec(prog, channels, tail, -1, -1);
#else
        rv = execlp(prog, channels, prog, tail, NULL);
        /* exec doesn't return job IDs, so no test needed */
#endif
    }
    itoa(rv, rval);
    return(rv == 0);
}

/*
 * Return a system dependant string with the current time.
 */

char *timeset(void)
{
    char    *sp;
    time_t  t;

    time(&t);
    sp = ctime(&t);
    sp[strlen(sp)-1] = 0;                   /* Lose \n at end */
    return(sp);
}

/*
 * This function accepts a full QDOS/SMS filename (device+path+name) and
 * finds the device, the path and the name of the file. Returns a QDOS error
 * code.
 */

int split_qdos_filename(char *fname, char *device, char *path, char *name)
{
    chanid_t    dumb;
    char        tmp[80], flag;
    qdirect_t   qlbuf;
    int         i, j, c, offset, ret;

    flag = TRUE;
    j    = 0;

    /*
     * Search for a device name in the form "devN_", "nM_devN_" or "nMM_devN_"
     * (where 0 < N < 9, "dev" is the device name, "n" is the network name, and
     * M/MM is the network address.
     */
    if (strlen(fname) > 8) {            /* First checks for a network name */
        if ((fname[2] == '_' && fname[7] == '_') ||
            (fname[3] == '_' && fname[8] == '_')) {

            if (!isalpha((unsigned char)fname[0]))
                flag = FALSE;

            if (!isdigit((unsigned char)fname[1]))
                flag = FALSE;

            if (fname[3] == '_') {
                j = 4;
                if (!isdigit((unsigned char)fname[2]))
                    flag = FALSE;
            } else
                j = 3;
        }
    }
    if (flag && strlen(fname) > 5) {    /* Now checks for the device name */
            for (i = j; i < j + 3; i++)
                if (!isalpha((unsigned char)fname[i]))
                    flag = FALSE;

            c = (int)(fname[j + 3] - '0');
            if (c < 1 || c > 8 || fname[j + 4] != '_')
                flag = FALSE;

            if (flag) {
                strncpy(device, fname, j + 5);
                device[j + 5] = '\0';
            } else
                return(ERR_BN);
    } else
        return(ERR_BN);

    strcpy(tmp, fname);
    j += 5;

    /*
     * Now search for the path and the file name...
     */
    while (strrchr(tmp, '_') != NULL) {
        offset = strrchr(tmp, '_') - tmp;
        strncpy(name, fname + (offset + 1), strlen(fname) - offset);
        name[strlen(fname) - offset] = '\0';

        if (strlen(fname) - strlen(name) - j > 0) {
            strncpy(path, fname + j, strlen(fname) - strlen(name) - j);
            path[strlen(fname) - strlen(name) - j] = '\0';
        } else {
            path[0] = '\0';
            break;
        }

        strcpy(tmp, device);
        strcat(tmp, path);

        if ((dumb = io_open(tmp, 1)) >= 0) {
            ret = fs_headr(dumb, -1, &qlbuf, 64);
            io_close(dumb);
            if (ret >= 0) {
                c = qlbuf.d_type;
                if (c == QF_DIR_TYPE || c == THOR_DIR_TYPE || c == CST_DIR_TYPE)
                    break;
            }
        }
        tmp[strlen(device) + strlen(path) - 1] = '\0';
    }
    return(0);
}

#if COMPLETION

/*
 * File directory routines. Do a directory search for file name completion.
 *
 * QDOS version written 7, 8/12/91 RJK.
 *
 * This code doesn't yet know about she .._ parent directory extension supplied
 * by C68 in other opens. Implementing it would be extreme hassle since it would
 * have to be put BACK into filenames returned.
 *
 * 29/5/92 Adapted to MicroEMACS 3.11.
 *
 * The only change necessary is the removal of the #if part - version 3.11
 * assumes that if no completion is available then the associated routines
 * always return NULL.
 *
 * 6/9/92 Rewrote the lot !
 *
 * Record the list of transformations needed to go from the given string to the
 * directory open name. When we get a filename, prepend the base device name
 * (=flp1_, NOT flp1_dir_) and then carry out transformations in reverse order.
 * Make sure everything is in lower case, since directory name case can be
 * dependent on filename.
 *
 * If an attempt to return a directory filename is made then it is returned once
 * with _ appended and shortly afterwards returned AGAIN with ._ appended. This
 * is so that you can't accidentally read directories as text files.
 */

struct dir_transform {
    struct  dir_transform *next;
    int     type;                       /* -ve for length of data dir,
                                           +ve for index to insert at. */
    char    text[1];                    /* Text for insertion begins here */
};

static struct dir_transform *dir_tlist; /* List of transforms */
static long   dir_channel = -1L;        /* Directory channel */
static char   dir_dirname[NFILEN];      /* Directory open name */
static short  dir_dlen;                 /* Length of device name */

/* Close and free directory data structures. Assumes very little. */

static void dir_tidy_up(void)
{
    struct dir_transform *t;

    if (dir_channel >= 0L) {
        io_close(dir_channel);
        dir_channel = -1L;
    }
    while (t = dir_tlist) {
        dir_tlist = dir_tlist->next;
        free(t);
    }
}

/*
 * Create a new directory name transform.
 * Depending on alignment restrictions, probably requests a byte or two
 * more than is required; but many allocations will round up to 8 or
 * even 16 bytes anyway, so who cares about the odd byte ?
 */

static struct dir_transform *new_transform(int extra_length)
{
    struct dir_transform *t;

    if (!(t = malloc((size_t)(sizeof(struct dir_transform) + extra_length)))) {
        dir_tidy_up();
        return(NULL);
    }
    t->next = dir_tlist;
    dir_tlist = t;
    return(t);
}

/* Get first filename. */

char *getffile(char *fname)
{
    int  extradata;
    char *p,*q;

    /* Clean up any left-overs */
    dir_tidy_up();

    /* Special cases */
    if (fname == NULL || *fname == '\0')
        return(NULL);

    /* Everything -> lower case */
    mklower(fname);

    /* Find full name. If name is a device name, it must be a directory device.
     * If default directory is null, then we can't work properly (this is
     * unlikely to ever happen, of course).
     */
    if (isdevice(fname, &extradata)) {
        if (!(extradata & DIRDEV))
            return(NULL);
        strcpy(dir_dirname, fname);
    } else {
        getcwd(dir_dirname, NFILEN);
        if (*dir_dirname) {
            if (!new_transform(0))
                return(NULL);
            dir_tlist->type = -strlen(dir_dirname);
            strcat(dir_dirname, fname);
        } else
            return(NULL);
    }

    /* Process "parent" and "self" directory specs. It is ASSUMED that strcpy()
     * start copying at the start of the string.
     */
    while (p = strstr(dir_dirname, "._")) {
        /* Can't work at start of name */
        if (p == dir_dirname) {
            dir_tidy_up();
            return(NULL);
        }
        if (p[-1] == '.') {
            /* Parent directory. Find underscore preceding dir name to delete;
             * if it isn't there then can't proceed.
             */
            q = p - 3;
            while (q >= dir_dirname && *q != '_')
                --q;
            if (q < dir_dirname) {
                dir_tidy_up();
                return(NULL);
            }
            /* We want to delete from q to p INCLUSIVE:-
             *
             * flp1_fred_joe_.._harry
             *          q     p
             *
             * e.g. delete "_joe_.." (= same as deleting "joe_.._").
             *
             * We must also remember allocate space for the \0 hence +2
             */
            if (!new_transform(p-q+2))
                return(NULL);
            dir_tlist->type = q - dir_dirname;
            strncpy(dir_tlist->text, q, p-q+1);
            dir_tlist->text[p-q+1] = '\0';
            strcpy(q, p+1);
        } else {
            /* Self directory. Just delete two characters, i.e. "._" */
            if (!new_transform(3))
                return(NULL);
            dir_tlist->type = p - dir_dirname;
            strcpy(dir_tlist->text, "._");
            strcpy(p, p + 2);
        }
    }

    /* Directory must contain an underscore after device name (we need it
     * because filenames in a directory don't have device names).
     */
    if (!(p = strchr(dir_dirname, '_'))) {
        dir_tidy_up();
        return(NULL);
    }
    dir_dlen = p - dir_dirname + 1;

    /* Try and open directory (at last). */
    if ((dir_channel = io_open(dir_dirname, 4)) < 0) {
        dir_tidy_up();
        return(NULL);
    }

    return(getnfile());
}

/*
 * Read the next file.
 * Directories are given special treatment so that you can't accidentally
 * load a directory as if it were an ordinary file (you can still do it
 * on purpose, though).
 *
 * 10/04/93: CST and THOR hard disk directories are recognized now.
 */

char *getnfile(void)
{
    struct qdirect fdef;                /* Directory entry */
    struct dir_transform *t;

    static char dir_flag;
    static char completename[NFILEN];

    if (dir_channel < 0L)               /* Shouldn't be needed, but... */
        return(NULL);

    if (dir_flag) {
        strcat(completename, "._");
        dir_flag = 0;
        return(completename);
    }

    if (dir_channel < 0L)               /* If not doing any reads, no files */
        return(NULL);

tryagain:
    /* Read the next filename */
    do {
        if (sizeof fdef != io_fstrg(dir_channel, -1, &fdef, sizeof fdef)) {
            dir_tidy_up();              /* End of file */
            return(NULL);
        }
    } while (!fdef.d_length);           /* Skip empty slots */

    /* Build filename */
    strncpy(completename, dir_dirname, dir_dlen);
    qlstr_to_c(completename + dir_dlen, (QLSTR_t *) &fdef.d_szname);

    /* Everything lower case */
    mklower(completename);

    /* Check initial substring */
    if (strncmp(completename, dir_dirname, strlen(dir_dirname)))
        goto tryagain;

    /* Special treatment for directory files (Applies only to level 2
     * drivers)
     */
    if (fdef.d_type == QF_DIR_TYPE || fdef.d_type == THOR_DIR_TYPE ||
                                                fdef.d_type == CST_DIR_TYPE) {
        dir_flag = 1;
        strcat(completename, "_");
    }

    /* Apply transformations */
    for (t = dir_tlist; t; t = t->next) {
        if (t->type < 0)
            strcpy(completename, completename - t->type);
        else {
            memmove(completename + t->type + strlen(t->text),
            completename + t->type,
            strlen(completename + t->type) + 1);
            strncpy(completename + t->type, t->text, strlen(t->text));
        }
    }

    return(completename);
}
#else
char *getffile(char *fspec)
{
    return(NULL);
}

char *getnfile(void)
{
    return(NULL);
}
#endif /* COMPLETION */

/*
 * qexec() function used as a _forkexec() bug work around
 */

#ifdef FORKEXECBUG

int qexec(char * fname, int * channels, char * cmd, jobid_t owner,
            timeout_t timeout)
{
    jobid_t         job_id;
    chanid_t        file_id, io_chan;
    qdirect_t       qlbuf;
    int             error, i;
    unsigned long   joblen, dataspace, cmdlen;
    char            *address, *stack;

    /* First try to open the program file */

    file_id = io_open(fname, 0);
    if (file_id < 0)
        return((int)file_id);

    /* Now check for the file type (must be EXECutable) */

    error = fs_headr(file_id, -1, &qlbuf, 64);
    if (error < 0 || qlbuf.d_type != 1) {
        io_close(file_id);
        if (error == 0)
            error = -15;
        return(error);
    }

    /* Calculate the job code length and the needed dataspace */

    cmdlen    = (strlen(cmd) + 3) & 0xfffffffe;
    joblen    = (qlbuf.d_length + 1) & 0xfffffffe;
    dataspace = (qlbuf.d_datalen + cmdlen + 9) & 0xfffffffe;
    if (channels != NULL)
        dataspace += channels[0] << 2;

    /* Create the job */

    job_id = mt_cjob(joblen, dataspace, NULL, owner, &address);
    if (job_id < 0) {
        io_close(file_id);
        return((int) job_id);
    }

    /* Load the program file into the job area */

    fs_pos(file_id, 0, 0);
    error = fs_load(file_id, address, qlbuf.d_length);
    io_close(file_id);
    if (error < 0) {
        mt_frjob(job_id, 0);
        return(error);
    }

    /* Put the environment variable pointer onto the stack */

    stack = address + joblen + dataspace - 4;
    *((long *)stack)  = (long) _VENV;
    stack -= 2;
    *((short *)stack) = ENV_MAGIC;

    /* Put the command line on the stack */

    stack -= cmdlen;
    cstr_to_ql((QLSTR_t *) stack, cmd);

    /* Put the number of channel Ids onto the stack */

    stack -= 2;
    if (channels != NULL) {
        i = channels[0];
        stack -= i << 2;
    } else
        i = 0;
    *((short *)stack) = i;

    /* Store the new stack pointer into the job header */

    address -= 12;
    *((long *)address) = (long) stack;

    /* Store the channel Id(s) onto the stack if any */

    if (channels != NULL) {
        stack += 2;
        for (i = 1; i <= channels[0]; i++) {
            file_id = getchid(channels[i]);
            *((long *)stack) = (long) file_id;
            stack += 4;
        }
    }

    /* Now activate the job */

    error = mt_activ(job_id, _def_priority, timeout);

    if (error != 0) {
        mt_frjob(job_id, 0);
        return(error);
    }

    /* Force a reschedule if the timeout is not infinite */

    if (timeout != -1)
        mt_susjb(-1, 5, NULL);

    /* Return without error */

    return(0);
}

#endif /* FORKEXECBUG */

#else
qdoshello() { }
#endif /* QDOS */
